from io import BytesIO
from pathlib import Path
from zipfile import BadZipFile

import logging
import warnings
from typing import Any, Hashable, Literal, overload
from warnings import warn

from pydantic import field_validator
from pydantic_core import ErrorDetails, ValidationError

from openpyxl import Workbook, load_workbook
import pandas as pd
import requests

from rdflib import Namespace


from cognite.neat.utils.auxiliary import local_import
from cognite.neat.constants import PREFIXES
from cognite.neat.rules import _exceptions
from cognite.neat.rules.models import Class, Metadata, Property, RuleModel, TransformationRules


@overload
def parse_rules_from_excel_file(filepath: Path, return_report: Literal[False] = False) -> TransformationRules:
    ...


@overload
def parse_rules_from_excel_file(
    filepath: Path, return_report: Literal[True]
) -> tuple[TransformationRules | None, list[ErrorDetails] | None, list | None]:
    ...


def parse_rules_from_excel_file(
    filepath: Path, return_report: bool = False
) -> tuple[TransformationRules | None, list[ErrorDetails] | None, list | None] | TransformationRules:
    """Parse transformation rules from an Excel file.

    Parameters
    ----------
    filepath : Path
        Path to the Excel file
    return_report : bool, optional
        Whether to return a report, by default False

    Returns
    -------
        Either the transformation rules or a tuple with the rules, a list of errors and a list of warnings
    """
    return from_tables(read_excel_file_to_table_by_name(filepath), return_report)


def parse_rules_from_google_sheet(
    sheet_id: str, return_report: bool = False
) -> tuple[TransformationRules | None, list[ErrorDetails] | None, list | None] | TransformationRules:
    """Parse transformation rules from a Google sheet.

    Parameters
    ----------
    sheet_id : str
        The identifier of the Google sheet with the rules.
    return_report : bool, optional
        Whether to return a report, by default False

    Returns
    -------
        Either the transformation rules or a tuple with the rules, a list of errors and a list of warnings
    """
    return from_tables(read_google_sheet_to_table_by_name(sheet_id), return_report)


def parse_rules_from_github_sheet(
    filepath: Path,
    personal_token: str,
    owner: str,
    repo: str,
    branch: str = "main",
    return_report: bool = False,
) -> tuple[TransformationRules | None, list[ErrorDetails] | None, list | None] | TransformationRules:
    """Parse transformation rules from a sheet stored in private Github.

    Parameters
    ----------
    filepath : Path
        Path to the sheet in the Github repository.
    personal_token : str
        Personal access token to access the Github repository.
    owner : str
        Owner of the Github repository.
    repo : str
        Name of the Github repository.
    branch : str, optional
        Branch of the Github repository, by default "main".

    Returns
    -------
        Either the transformation rules or a tuple with the rules, a list of errors and a list of warnings
    """

    return from_tables(read_github_sheet_to_table_by_name(filepath, personal_token, owner, repo, branch), return_report)


def parse_rules_from_yaml(dirpath: Path) -> TransformationRules:
    """
    Load transformation rules from a yaml file.

    Args:
        dirpath (Path): Path to the yaml file.
    Returns:
        TransformationRules: The transformation rules.
    """
    return TransformationRules(**read_yaml_file_to_mapping_by_name(dirpath))


def from_tables(
    raw_dfs: dict[str, pd.DataFrame], return_report: bool = False
) -> tuple[TransformationRules | None, list[ErrorDetails] | None, list | None] | TransformationRules:
    # the only way to suppress warnings from pylense
    validation_warnings = []
    try:
        with warnings.catch_warnings(record=True) as validation_warnings:
            raw_tables = RawTables.from_raw_dataframes(raw_dfs)
            rules_dict: dict[str, Any] = {
                "metadata": _parse_metadata(raw_tables.Metadata),
                "classes": _parse_classes(raw_tables.Classes),
                "properties": _parse_properties(raw_tables.Properties),
                "prefixes": PREFIXES if raw_tables.Prefixes.empty else _parse_prefixes(raw_tables.Prefixes),
            }

            rules_dict["instances"] = (
                None
                if raw_tables.Instances.empty
                else _parse_instances(raw_tables.Instances, rules_dict["metadata"], rules_dict["prefixes"])
            )
            rules = TransformationRules(**rules_dict)
        return (rules, None, _exceptions.wrangle_warnings(validation_warnings)) if return_report else rules

    except _exceptions.Error0 as e:
        validation_errors = [e.to_error_dict()]
        if return_report:
            return None, validation_errors, _exceptions.wrangle_warnings(validation_warnings)
        else:
            raise e
    except ValidationError as e:
        validation_errors = e.errors()
        if return_report:
            return None, validation_errors, _exceptions.wrangle_warnings(validation_warnings)
        else:
            raise e


def _parse_metadata(meta_df: pd.DataFrame) -> dict[str, Any]:
    metadata_dict = dict(zip(meta_df[0], meta_df[1]))
    metadata_dict["source"] = meta_df.source if "source" in dir(meta_df) else None
    if "namespace" in metadata_dict:
        metadata_dict["namespace"] = Namespace(metadata_dict["namespace"])
    return metadata_dict


def _parse_classes(classes_df: pd.DataFrame) -> dict[Any | None, dict[Hashable, Any]]:
    return {class_.get("Class"): class_ for class_ in classes_df.to_dict(orient="records")}


def _parse_properties(properties_df: pd.DataFrame) -> dict[str, dict[Hashable, Any]]:
    return {f"row {i+3}": property_ for i, property_ in enumerate(properties_df.to_dict(orient="records"))}


def _parse_prefixes(prefix_df: pd.DataFrame) -> dict[str, Namespace]:
    return {row["Prefix"]: Namespace(row["URI"]) for i, row in prefix_df.iterrows()}


def _parse_instances(
    instances_df: pd.DataFrame, metadata: dict[str, Any], prefixes: dict[str, Namespace]
) -> list[dict] | None:
    if "prefix" not in metadata or "namespace" not in metadata:
        logging.warning(_exceptions.Warning500().message)
        warn(_exceptions.Warning500().message)
        return None

    prefixes[metadata["prefix"]] = metadata["namespace"]

    instances = []
    for _, row in instances_df.iterrows():
        row_as_dict = row.to_dict()
        row_as_dict["namespace"] = metadata["namespace"]
        row_as_dict["prefixes"] = prefixes
        instances.append(row_as_dict)
    return instances


class RawTables(RuleModel):
    Metadata: pd.DataFrame
    Properties: pd.DataFrame
    Classes: pd.DataFrame
    Prefixes: pd.DataFrame = pd.DataFrame()
    Instances: pd.DataFrame = pd.DataFrame()

    @classmethod
    def from_raw_dataframes(cls, raw_dfs: dict[str, pd.DataFrame]) -> "RawTables":
        expected_tables = cls.mandatory_fields()

        # Validate raw tables
        if missing_tables := (expected_tables - set(raw_dfs)):
            raise _exceptions.Error0(missing_tables)

        tables_dict = {
            Tables.metadata: raw_dfs[Tables.metadata],
            Tables.classes: raw_dfs[Tables.classes],
            Tables.properties: raw_dfs[Tables.properties],
        }

        if Tables.prefixes in raw_dfs:
            tables_dict[Tables.prefixes] = raw_dfs[Tables.prefixes]
        if Tables.instances in raw_dfs:
            tables_dict[Tables.instances] = raw_dfs[Tables.instances]

        return cls(**tables_dict)

    @field_validator("Metadata")
    def has_metadata_mandatory_rows(cls, v):
        given_rows = set(v[0].values)
        mandatory_rows = Metadata.mandatory_fields()
        mandatory_rows_alias = Metadata.mandatory_fields(use_alias=True)

        if not (mandatory_rows.issubset(given_rows) or mandatory_rows_alias.issubset(given_rows)):
            missing_rows = mandatory_rows_alias.difference(given_rows)
            raise _exceptions.Error51(missing_rows).to_pydantic_custom_error()
        return v

    @field_validator("Classes")
    def has_classes_mandatory_columns(cls, v):
        given_columns = set(v.columns)
        mandatory_columns = Class.mandatory_fields()
        mandatory_columns_alias = Class.mandatory_fields(use_alias=True)

        if not (mandatory_columns.issubset(given_columns) or mandatory_columns_alias.issubset(given_columns)):
            missing_columns = mandatory_columns_alias.difference(given_columns)
            raise _exceptions.Error52(missing_columns).to_pydantic_custom_error()
        return v

    @field_validator("Properties")
    def has_properties_mandatory_columns(cls, v):
        given_columns = set(v.columns)
        mandatory_columns = Property.mandatory_fields()
        mandatory_columns_alias = Property.mandatory_fields(use_alias=True)

        if not (mandatory_columns.issubset(given_columns) or mandatory_columns_alias.issubset(given_columns)):
            missing_columns = mandatory_columns_alias.difference(given_columns)
            raise _exceptions.Error53(missing_columns).to_pydantic_custom_error()
        return v

    @field_validator("Prefixes")
    def has_prefixes_mandatory_columns(cls, v):
        given_columns = set(v.columns)
        mandatory_columns = {"Prefix", "URI"}

        if not mandatory_columns.issubset(given_columns):
            missing_columns = mandatory_columns.difference(given_columns)
            raise _exceptions.Error54(missing_columns).to_pydantic_custom_error()
        return v

    @field_validator("Instances")
    def has_instances_mandatory_columns(cls, v):
        given_columns = set(v.columns)
        mandatory_columns = {"Instance", "Property", "Value"}

        if not mandatory_columns.issubset(given_columns):
            missing_columns = mandatory_columns.difference(given_columns)
            raise _exceptions.Error55(missing_columns).to_pydantic_custom_error()
        return v


class Tables:
    prefixes = "Prefixes"
    properties = "Properties"
    classes = "Classes"
    metadata = "Metadata"
    instances = "Instances"


# readers:


def read_google_sheet_to_table_by_name(sheet_id: str) -> dict[str, pd.DataFrame]:
    # To trigger ImportError if gspread is not installed
    local_import("gspread", "google")
    import gspread

    client_google = gspread.service_account()
    spreadsheet = client_google.open_by_key(sheet_id)
    return {worksheet.title: pd.DataFrame(worksheet.get_all_records()) for worksheet in spreadsheet.worksheets()}


def read_excel_file_to_table_by_name(filepath: Path) -> dict[str, pd.DataFrame]:
    # To trigger ImportError if openpyxl is not installed
    local_import("openpyxl", "excel")

    from openpyxl import Workbook, load_workbook

    workbook: Workbook = load_workbook(filepath)

    sheets = {
        sheetname: pd.read_excel(
            filepath,
            sheet_name=sheetname,
            header=None if sheetname == "Metadata" else 0,
            skiprows=1 if sheetname in ["Classes", "Properties", "Instances"] else None,
        )
        for sheetname in workbook.sheetnames
    }

    for sheetname in sheets:
        sheets[sheetname].source = filepath

    return sheets


def read_yaml_file_to_mapping_by_name(dirpath: Path, expected_files: set[str] | None = None) -> dict[str, dict]:
    # To trigger ImportError if yaml is not installed
    local_import("yaml", "yaml")
    from yaml import safe_load

    mapping_by_name = {}
    for filepath in dirpath.iterdir():
        if expected_files is not None and filepath.stem not in expected_files:
            continue
        mapping_by_name[filepath.stem] = safe_load(filepath.read_text())
    return mapping_by_name


def read_github_sheet_to_table_by_name(
    filepath: str, personal_token: str, owner: str, repo: str, branch: str = "main"
) -> dict[str, pd.DataFrame]:
    r = requests.get(
        f"https://api.github.com/repos/{owner}/{repo}/contents/{filepath}?ref={branch}",
        headers={"accept": "application/vnd.github.v3.raw", "authorization": f"token {personal_token}"},
    )

    loc = f"https://github.com/{owner}/{repo}/tree/{branch}"

    if r.status_code != 200:
        raise _exceptions.Error20(filepath, loc, r.reason)
    try:
        wb = load_workbook(BytesIO(r.content), data_only=True)
    except BadZipFile:
        raise _exceptions.Error21(filepath, loc)
    return _workbook_to_table_by_name(wb)


def _workbook_to_table_by_name(workbook: Workbook) -> dict[str, pd.DataFrame]:
    table = {}
    for sheet in workbook:
        sheetname = sheet.title
        data = sheet.values
        if sheetname == "Metadata":
            table[sheetname] = pd.DataFrame(data, columns=None)
        if sheetname in ["Classes", "Properties", "Instances"]:
            next(data)
            columns = next(data)[:]
            table[sheet.title] = pd.DataFrame(data, columns=columns).dropna(how="all")

    return table



import logging
import time
import traceback
from fastapi import APIRouter
import rdflib

from cognite.neat.graph.transformations import query_generator

from cognite.neat.app.api.configuration import neat_app, cache_store
from cognite.neat.app.api.data_classes.rest import NodesAndEdgesRequest, QueryRequest, RuleRequest
from cognite.neat.app.api.asgi.metrics import counter
from cognite.neat.app.api.utils.data_mapping import rdf_result_to_api_response
from cognite.neat.app.api.utils.query_templates import query_templates


router = APIRouter()


@router.get("/api/list-queries")
def list_queries():
    counter.inc()
    return query_templates


@router.post("/api/get-nodes-and-edges")
def get_nodes_and_edges(request: NodesAndEdgesRequest):
    logging.info("Querying graph nodes and edges :")
    if "get_nodes_and_edges" in cache_store and request.cache:
        return cache_store["get_nodes_and_edges"]
    nodes_result = {}
    edges_result = {}
    query = ""
    elapsed_time_sec = 0
    """
    "nodes": [
        {
            "node_id": "http://purl.org/nordic44#_a63ce14e-fba0-4f9e-8b59-7ef2fe887ff8",
            "node_class": "http://entsoe.eu/CIM/SchemaExtension/3/2#RateTemperature",
            "node_name": "-5degreeCelsius"
        }],
    "edges": [
        {
            "src_object_ref": "http://purl.org/nordic44#_f1769eaa-9aeb-11e5-91da-b8763fd99c5f",
            "conn": "http://iec.ch/TC57/2013/CIM-schema-cim16#OperationalLimitSet.Terminal",
            "dst_object_ref": "http://purl.org/nordic44#_2dd90186-bdfb-11e5-94fa-c8f73332c8f4"
        },
    """
    if request.sparql_query:
        mixed_result = get_data_from_graph(request.sparql_query, request.graph_name, request.workflow_name)

        edges_result = []
        nodes_result = []
        try:
            nodes_result = [
                {
                    "node_id": v[rdflib.Variable("node_id")],
                    "node_class": v[rdflib.Variable("node_class")],
                    "node_name": v[rdflib.Variable("node_name")],
                }
                for v in mixed_result["rows"]
            ]
        except Exception as e:
            logging.error(f"Error while parsing nodes : {e}")

        try:
            edges_result = [
                {
                    "src_object_ref": v[rdflib.Variable("src_object_ref")],
                    "dst_object_ref": v[rdflib.Variable("dst_object_ref")],
                }
                for v in mixed_result["rows"]
                if rdflib.Variable("src_object_ref") in v and rdflib.Variable("dst_object_ref") in v
            ]
        except Exception as e:
            logging.error(f"Error while parsing edges : {e}")

        query = request.sparql_query
        elapsed_time_sec = mixed_result["elapsed_time_sec"]
    else:
        nodes_filter = ""
        edges_dst_filter = ""
        edges_src_filter = ""

        if len(request.node_class_filter) > 0:
            nodes_filter = "VALUES ?node_class { " + " ".join([f"<{x}>" for x in request.node_class_filter]) + " }"

        node_name_property = request.node_name_property or "cim:IdentifiedObject.name"
        nodes_query = f"SELECT DISTINCT ?node_id ?node_class ?node_name WHERE \
        {{ {nodes_filter} \
        ?node_id {node_name_property} ?node_name . ?node_id rdf:type ?node_class }} "
        if len(request.src_edge_filter) > 0:
            edges_src_filter = (
                "VALUES ?src_object_class { " + " ".join([f"<{x}>" for x in request.src_edge_filter]) + " }"
            )
        if len(request.dst_edge_filter) > 0:
            edges_dst_filter = (
                "VALUES ?dst_object_class { " + " ".join([f"<{x}>" for x in request.dst_edge_filter]) + " }"
            )

        edges_query = f"SELECT  ?src_object_ref ?conn ?dst_object_ref \
            WHERE {{ \
            {edges_src_filter} \
            {edges_dst_filter} \
            ?src_object_ref ?conn  ?dst_object_ref . \
            ?src_object_ref  rdf:type ?src_object_class . \
            ?dst_object_ref  rdf:type ?dst_object_class .\
            }} "
        nodes_query = f"{nodes_query} LIMIT {request.limit}"
        edges_query = f"{edges_query} LIMIT {request.limit}"
        logging.info(f"Nodes query : {nodes_query}")
        logging.info(f"Edges query : {edges_query}")
        nodes_result = get_data_from_graph(nodes_query, request.graph_name, request.workflow_name)
        edges_result = get_data_from_graph(edges_query, request.graph_name, request.workflow_name)
        elapsed_time_sec = nodes_result["elapsed_time_sec"] + edges_result["elapsed_time_sec"]
        query = nodes_query + edges_query
        nodes_result = nodes_result["rows"]
        edges_result = edges_result["rows"]

    merged_result = {
        "nodes": nodes_result,
        "edges": edges_result,
        "error": "",
        "elapsed_time_sec": elapsed_time_sec,
        "query": query,
    }

    if request.cache:
        cache_store["get_nodes_and_edges"] = merged_result
    return merged_result


@router.post("/api/query")
def query_graph(request: QueryRequest):
    logging.info(f"Querying graph {request.graph_name} with query {request.query}")
    sparq_query = request.query
    return get_data_from_graph(sparq_query, request.graph_name, workflow_name=request.workflow_name)


@router.post("/api/execute-rule")
def execute_rule(request: RuleRequest):
    logging.debug(
        f"Executing rule type: { request.rule_type } rule : {request.rule} , workflow : {request.workflow_name} , "
        f"graph : {request.graph_name}"
    )
    workflow = neat_app.workflow_manager.get_workflow(request.workflow_name)

    api_result = {"error": ""}
    workflow_context = workflow.get_context()
    if request.graph_name == "source":
        if "SourceGraph" in workflow_context:
            graph = workflow_context["SourceGraph"].graph
        else:
            logging.info("Source graph is empty , please load the graph first")
            api_result["error"] = "source graph is empty , please load the graph first"
    elif request.graph_name == "solution":
        if "SolutionGraph" in workflow_context:
            graph = workflow_context["SolutionGraph"].graph
        else:
            logging.info("Solution graph is empty , please load the graph first")
            api_result["error"] = "solution graph is empty , please load the graph first"
    else:
        raise Exception("Unknown graph name")

    if request.rule_type == "rdfpath":
        start_time = time.perf_counter()
        sparq_query = query_generator.build_sparql_query(
            graph, request.rule, prefixes=workflow_context["RulesData"].rules.prefixes
        )
    else:
        logging.error("Unknown rule type")
        return {"error": "Unknown rule type"}
    stop_time = time.perf_counter()
    elapsed_time_sec_1 = stop_time - start_time
    logging.info(f"Computed query : {sparq_query} in {elapsed_time_sec_1 * 1000} ms")

    api_result = get_data_from_graph(sparq_query, request.graph_name, workflow_name=request.workflow_name)
    api_result["elapsed_time_sec"] += elapsed_time_sec_1
    return api_result


@router.get("/api/object-properties")
def get_object_properties(reference: str, graph_name: str = "source", workflow_name: str = "default"):
    logging.info(f"Querying object-properties from {graph_name} :")
    query = f"SELECT ?property ?value WHERE {{ <{reference}> ?property ?value }} ORDER BY ?property"
    return get_data_from_graph(query, graph_name, workflow_name=workflow_name)


@router.get("/api/search")
def search(
    search_str: str, graph_name: str = "source", search_type: str = "value_exact_match", workflow_name: str = "default"
):
    logging.info(f"Search from {graph_name} :")
    if search_type == "reference":
        query = (
            f"SELECT ?class ?reference WHERE {{ {{?reference ?class <{search_str}> . }} "
            f"UNION {{ <{search_str}> ?class ?reference }} }} limit 10000"
        )

    elif search_type == "value_exact_match":
        query = (
            f"select ?object_ref ?type ?property ?value where "
            f"{{ ?object_ref ?property ?value . ?object_ref rdf:type ?type . "
            f'FILTER(?value="{search_str}") }} limit 10000'
        )
    elif search_type == "value_regexp":
        query = (
            f"select ?object_ref ?type ?property ?value where "
            f"{{ ?object_ref ?property ?value . ?object_ref rdf:type ?type .  "
            f'FILTER regex(?value, "{search_str}","i") }} limit 10000'
        )
    return get_data_from_graph(query, graph_name, workflow_name=workflow_name)


@router.get("/api/get-classes")
def get_classes(graph_name: str = "source", workflow_name: str = "default", cache: bool = True):
    logging.info(f"Querying graph classes from graph name : {graph_name}, cache : {cache}")
    cache_key = f"classes_{graph_name}"
    if cache_key in cache_store and cache:
        return cache_store[cache_key]
    query = (
        "SELECT ?class (count(?s) as ?instances ) WHERE { ?s rdf:type ?class . } "
        "group by ?class order by DESC(?instances)"
    )
    api_result = get_data_from_graph(query, graph_name, workflow_name)
    cache_store["get_classes"] = api_result
    return api_result


def get_data_from_graph(sparq_query: str, graph_name: str = "source", workflow_name: str = "default"):
    total_elapsed_time = 0
    api_result = {"error": ""}
    result = None

    try:
        logging.info(f"Preparing query :{sparq_query} ")
        start_time = time.perf_counter()
        workflow = neat_app.workflow_manager.get_workflow(workflow_name)
        workflow_context = workflow.get_context()

        if graph_name == "source":
            if "SourceGraph" in workflow_context:
                result = workflow_context["SourceGraph"].graph.query(sparq_query)
            else:
                logging.info("Source graph is empty , please load the graph first")
                api_result["error"] = "source graph is empty , please load the graph first"
        elif graph_name == "solution":
            if "SolutionGraph" in workflow_context:
                result = workflow_context["SolutionGraph"].graph.query(sparq_query)
            else:
                logging.info("Solution graph is empty , please load the graph first")
                api_result["error"] = "solution graph is empty , please load the graph first"
        else:
            raise Exception("Unknown graph name")

        stop_time = time.perf_counter()
        elapsed_time_sec_1 = stop_time - start_time
        logging.info(f"Query prepared in {elapsed_time_sec_1 * 1000} ms")

        start_time = time.perf_counter()
        if result:
            api_result = rdf_result_to_api_response(result)
        stop_time = time.perf_counter()
        elapsed_time_sec_2 = stop_time - start_time
        total_elapsed_time = elapsed_time_sec_1 + elapsed_time_sec_2
    except Exception as e:
        logging.error(f"Error while executing query :{e}")
        traceback.print_exc()
        api_result["error"] = str(e)

    api_result["query"] = sparq_query
    api_result["elapsed_time_sec"] = total_elapsed_time
    logging.info(f"Data fetched in {total_elapsed_time * 1000} ms")
    return api_result



from cognite.neat.rules.models import TransformationRules


def get_labels(transformation_rules: TransformationRules) -> set[str]:
    """Return CDF labels for classes and relationships."""
    class_labels = {class_.class_id for class_ in transformation_rules.classes.values()}

    property_labels = {property_.property_id for property_ in transformation_rules.properties.values()}

    relationship_labels = {
        rule.label for rule in transformation_rules.properties.values() if "Relationship" in rule.cdf_resource_type
    }

    return class_labels.union(relationship_labels).union(property_labels)



"""Exports rules to CDF Data Model Storage (DMS) through cognite-sdk.
"""

import logging
from typing import ClassVar, Optional, Self
import warnings
from pydantic import BaseModel, ConfigDict

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import ContainerApply, ContainerProperty, DirectRelation
from cognite.client.data_classes.data_modeling import (
    ViewApply,
    SpaceApply,
    DataModelApply,
    DirectRelationReference,
)
from cognite.client.data_classes.data_modeling import (
    MappedPropertyApply,
    ContainerId,
    ViewId,
    SingleHopConnectionDefinition,
)
from cognite.neat.rules.analysis import to_class_property_pairs

from cognite.neat.rules.models import Property, TransformationRules, DATA_TYPE_MAPPING
from cognite.neat.rules import _exceptions
from cognite.neat.rules._validation import (
    are_entity_names_dms_compliant,
    are_properties_redefined,
)
from cognite.neat.utils.utils import generate_exception_report


class DataModel(BaseModel):
    """Data model pydantic class used to create space, containers, views and data model in CDF.
    based on the transformation rules."""

    space: str
    external_id: str
    version: str
    description: Optional[str] = None
    name: Optional[str] = None
    containers: dict[str, ContainerApply]
    views: dict[str, ViewApply]

    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )

    @classmethod
    def from_rules(cls, transformation_rules: TransformationRules) -> Self:
        """Generates a DataModel class instance from a TransformationRules instance.

        Args:
            transformation_rules (TransformationRules): instance of TransformationRules.

        Returns:
            DataModel: instance of DataModel.
        """
        names_compliant, name_warnings = are_entity_names_dms_compliant(transformation_rules, return_report=True)
        if not names_compliant:
            logging.error(_exceptions.Error10(report=generate_exception_report(name_warnings)).message)
            raise _exceptions.Error10(report=generate_exception_report(name_warnings))

        properties_redefined, redefinition_warnings = are_properties_redefined(transformation_rules, return_report=True)
        if properties_redefined:
            logging.error(_exceptions.Error11(report=generate_exception_report(redefinition_warnings)))
            raise _exceptions.Error11(report=generate_exception_report(redefinition_warnings))

        return cls(
            space=transformation_rules.metadata.cdf_space_name,
            external_id=transformation_rules.metadata.data_model_name,
            version=transformation_rules.metadata.version,
            description=transformation_rules.metadata.description,
            name=transformation_rules.metadata.title,
            containers=cls.containers_from_rules(transformation_rules),
            views=cls.views_from_rules(transformation_rules),
        )

    @staticmethod
    def containers_from_rules(transformation_rules: TransformationRules) -> dict[str, ContainerApply]:
        """Create a dictionary of ContainerApply instances from a TransformationRules instance.

        Args:
            transformation_rules (TransformationRules): instance of TransformationRules.`

        Returns:
            dict[str, ContainerApply]: dictionary of ContainerApply instances.
        """
        class_properties = to_class_property_pairs(transformation_rules)
        return {
            class_id: ContainerApply(
                space=transformation_rules.metadata.cdf_space_name,
                external_id=class_id,
                name=transformation_rules.classes[class_id].class_name,
                description=transformation_rules.classes[class_id].description,
                properties=DataModel.container_properties_from_dict(properties),
            )
            for class_id, properties in class_properties.items()
        }

    @staticmethod
    def container_properties_from_dict(properties: dict[str, Property]) -> dict[str, ContainerProperty]:
        """Generates a dictionary of ContainerProperty instances from a dictionary of Property instances.

        Args:
            properties (dict[str, Property]): dictionary of Property instances.

        Returns:
            dict[str, ContainerProperty]: dictionary of ContainerProperty instances.
        """
        container_properties = {}
        for property_id, property_definition in properties.items():
            # Literal, i.e. attribute
            if property_definition.property_type == "DatatypeProperty":
                container_properties[property_id] = ContainerProperty(
                    type=DATA_TYPE_MAPPING[property_definition.expected_value_type]["dms"](
                        is_list=property_definition.max_count != 1
                    ),
                    nullable=property_definition.min_count == 0,
                    default_value=property_definition.default,
                    name=property_definition.property_name,
                    description=property_definition.description,
                )

            # URIRef, i.e. edge
            elif property_definition.property_type == "ObjectProperty":
                container_properties[property_id] = ContainerProperty(
                    type=DirectRelation(),
                    nullable=True,
                    name=property_definition.property_name,
                    description=property_definition.description,
                )
            else:
                logging.warning(_exceptions.Warning60(property_id, property_definition.property_type).message)
                warnings.warn(
                    _exceptions.Warning60(property_id, property_definition.property_type).message,
                    category=_exceptions.Warning60,
                    stacklevel=2,
                )

        return container_properties

    @staticmethod
    def views_from_rules(transformation_rules: TransformationRules) -> dict[str, ViewApply]:
        """Generates a dictionary of ViewApply instances from a TransformationRules instance.

        Args:
            transformation_rules (TransformationRules): instance of TransformationRules.

        Returns:
            dict[str, ViewApply]: dictionary of ViewApply instances.
        """
        class_properties = to_class_property_pairs(transformation_rules)
        return {
            class_id: ViewApply(
                space=transformation_rules.metadata.cdf_space_name,
                external_id=class_id,
                name=transformation_rules.classes[class_id].class_name,
                description=transformation_rules.classes[class_id].description,
                properties=DataModel.view_properties_from_dict(
                    properties, transformation_rules.metadata.cdf_space_name, transformation_rules.metadata.version
                ),
                version=transformation_rules.metadata.version,
            )
            for class_id, properties in class_properties.items()
        }

    @staticmethod
    def view_properties_from_dict(
        properties: dict[str, Property], space: str, version: str
    ) -> dict[str, MappedPropertyApply | SingleHopConnectionDefinition]:
        view_properties = {}
        for property_id, property_definition in properties.items():
            # attribute
            if property_definition.property_type == "DatatypeProperty":
                view_properties[property_id] = MappedPropertyApply(
                    container=ContainerId(space=space, external_id=property_definition.class_id),
                    container_property_identifier=property_id,
                    name=property_definition.property_name,
                    description=property_definition.description,
                )

            # edge 1-1
            elif property_definition.property_type == "ObjectProperty" and property_definition.max_count == 1:
                view_properties[property_id] = MappedPropertyApply(
                    container=ContainerId(space=space, external_id=property_definition.class_id),
                    container_property_identifier=property_id,
                    name=property_definition.property_name,
                    description=property_definition.description,
                    source=ViewId(space=space, external_id=property_definition.expected_value_type, version=version),
                )

            # edge 1-many
            elif property_definition.property_type == "ObjectProperty" and property_definition.max_count != 1:
                view_properties[property_id] = SingleHopConnectionDefinition(
                    type=DirectRelationReference(
                        space=space, external_id=f"{property_definition.class_id}.{property_definition.property_id}"
                    ),
                    source=ViewId(space=space, external_id=property_definition.expected_value_type, version=version),
                    direction="outwards",
                    name=property_definition.property_name,
                    description=property_definition.description,
                )
            else:
                logging.warning(_exceptions.Warning61(property_id).message)
                warnings.warn(
                    _exceptions.Warning61(property_id).message,
                    category=_exceptions.Warning61,
                    stacklevel=2,
                )

        return view_properties

    def to_cdf(self, client: CogniteClient):
        """Creates the data model in CDF.

        Args:
            client (CogniteClient): Cognite client.
        """
        existing_data_model = self.find_existing_data_model(client)
        existing_containers = self.find_existing_containers(client)
        existing_views = self.find_existing_views(client)

        if existing_data_model or existing_containers or existing_views:
            raise _exceptions.Error60(existing_data_model, existing_containers, existing_views)

        self.create_space(client)
        self.create_containers(client)
        self.create_views(client)
        self.create_data_model(client)

    def find_existing_data_model(self, client: CogniteClient) -> set[str]:
        """Checks if the data model exists in CDF.

        Args:
            client (CogniteClient): Cognite client.

        Returns:
            bool: True if the data model exists, False otherwise.
        """

        cdf_data_model = {}

        if model := client.data_modeling.data_models.retrieve((self.space, self.external_id, self.version)):
            cdf_data_model = model[0]
            logging.warning(_exceptions.Warning64(self.external_id, self.version, self.space).message)
            warnings.warn(
                _exceptions.Warning64(self.external_id, self.version, self.space).message,
                category=_exceptions.Warning64,
                stacklevel=2,
            )

            return {cdf_data_model.external_id}
        else:
            return set()

    def find_existing_containers(self, client: CogniteClient) -> set[str]:
        """Checks if the containers exist in CDF.

        Args:
            client (CogniteClient): Cognite client.

        Returns:
            bool: True if the containers exist, False otherwise.
        """

        cdf_containers = {}
        if containers := client.data_modeling.containers.list(space=self.space, limit=-1):
            cdf_containers = {container.external_id: container for container in containers}

        if existing_containers := set(self.containers.keys()).intersection(set(cdf_containers.keys())):
            logging.warning(_exceptions.Warning62(existing_containers, self.space).message)
            warnings.warn(
                _exceptions.Warning62(existing_containers, self.space).message,
                category=_exceptions.Warning62,
                stacklevel=2,
            )

            return existing_containers
        else:
            return set()

    def find_existing_views(self, client: CogniteClient) -> set[str]:
        """Checks if the views exist in CDF.

        Args:
            client (CogniteClient): Cognite client.

        Returns:
            bool: True if the views exist, False otherwise.
        """
        cdf_views = {}
        if views := client.data_modeling.views.list(space=self.space, limit=-1):
            cdf_views = {view.external_id: view for view in views if view.version == self.version}

        if existing_views := set(self.views.keys()).intersection(set(cdf_views.keys())):
            logging.warning(_exceptions.Warning63(existing_views, self.version, self.space).message)
            warnings.warn(
                _exceptions.Warning63(existing_views, self.version, self.space).message,
                category=_exceptions.Warning63,
                stacklevel=2,
            )
            return existing_views
        else:
            return set()

    def create_space(self, client: CogniteClient):
        logging.info(f"Creating space {self.space}")
        _ = client.data_modeling.spaces.apply(SpaceApply(space=self.space))

    def create_containers(self, client: CogniteClient):
        for container_id in self.containers:
            logging.info(f"Creating container {container_id} in space {self.space}")
            _ = client.data_modeling.containers.apply(self.containers[container_id])

    def create_views(self, client: CogniteClient):
        for view_id in self.views:
            logging.info(f"Creating view {view_id} version {self.views[view_id].version} in space {self.space}")
            _ = client.data_modeling.views.apply(self.views[view_id])

    def create_data_model(self, client: CogniteClient):
        logging.info(f"Creating data model {self.external_id} version {self.version} in space {self.space}")
        _ = client.data_modeling.data_models.apply(
            DataModelApply(
                name=self.name,
                description=self.description,
                space=self.space,
                external_id=self.external_id,
                version=self.version,
                views=list(self.views.values()),
            )
        )

    def remove_data_model(self, client: CogniteClient):
        """Helper function to remove a data model, and all underlying views and containers from CDF.

        Args:
            client (CogniteClient): Cognite client.
        """

        if client.data_modeling.data_models.retrieve((self.space, self.external_id, self.version)):
            logging.info(f"Removing data model {self.external_id} version {self.version} from space {self.space}")
            _ = client.data_modeling.data_models.delete((self.space, self.external_id, self.version))

        if views := client.data_modeling.views.retrieve(
            [view.as_id() for view in self.views.values()], all_versions=False
        ):
            for view in views:
                logging.info(f"Removing view {view.external_id} version {view.version} from space {self.space}")
                _ = client.data_modeling.views.delete((view.space, view.external_id, view.version))

        if containers := client.data_modeling.containers.retrieve(
            [container.as_id() for container in self.containers.values()]
        ):
            for container in containers:
                logging.info(f"Removing container {container.external_id} from space {self.space}")
                _ = client.data_modeling.containers.delete((container.space, container.external_id))



import logging
import warnings

# import traceback
from typing import Dict, List, Set, Union
from warnings import warn

import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import LabelFilter, Relationship, RelationshipUpdate
from cognite.client.exceptions import CogniteDuplicatedError

from cognite.neat.graph.loaders.core.models import RelationshipDefinition, RelationshipDefinitions
from cognite.neat.graph.loaders.core.rdf_to_assets import _categorize_cdf_assets
from cognite.neat.graph.stores import NeatGraphStore
from cognite.neat.rules.models import TransformationRules
from cognite.neat.utils.utils import chunker, datetime_utc_now, epoch_now_ms, remove_namespace, retry_decorator


def define_relationships(rules: TransformationRules, stop_on_exception: bool = False) -> RelationshipDefinitions:
    relationships = {}

    # Unique ids used to check for redefinitions of relationships
    ids = set()

    for row, rule in rules.properties.items():
        if "Relationship" in rule.cdf_resource_type:
            relationship = RelationshipDefinition(
                source_class=rule.class_id,
                target_class=rule.expected_value_type,
                property_=rule.property_id,
                labels=list(
                    set([rule.label, rule.class_id, rule.expected_value_type, "non-historic", rule.property_id])
                ),
                target_type=rule.target_type,
                source_type=rule.source_type,
                relationship_external_id_rule=rule.relationship_external_id_rule,
            )

            id_ = f"{rule.class_id}({rule.property_id})"
            if id_ in ids:
                msg = f"Relationship {rule.property_id} redefined at {row} in transformation rules!"
                if stop_on_exception:
                    logging.error(msg)
                    raise ValueError(msg)
                else:
                    msg += " Skipping redefinition!"
                    warnings.warn(msg, stacklevel=2)
                    logging.warning(msg)
            else:
                relationships[row] = relationship
                ids.add(id_)

    if relationships:
        return RelationshipDefinitions(
            data_set_id=rules.metadata.data_set_id,
            prefix=rules.metadata.prefix,
            namespace=rules.metadata.namespace,
            relationships=relationships,
        )

    msg = "No relationship defined in transformation rule sheet!"
    if stop_on_exception:
        logging.error(msg)
        raise ValueError(msg)
    else:
        warnings.warn(msg, stacklevel=2)
        logging.warning(msg)
        return RelationshipDefinitions(
            data_set_id=rules.metadata.data_set_id,
            prefix=rules.metadata.prefix,
            namespace=rules.metadata.namespace,
            relationships={},
        )


# should be renamed to rdf2relationship_data_frame -> rdf2relationships
def rdf2relationships(
    graph_store: NeatGraphStore,
    transformation_rules: TransformationRules,
    stop_on_exception: bool = False,
) -> pd.DataFrame:
    """_summary_

    Parameters
    ----------
    graph : Graph
        _description_
    transformation_rules : TransformationRules
        _description_

    Returns
    -------
    Dict[str, Asset]
        _description_
    """

    # Step 1: Generate relationship definitions
    relationship_definitions = define_relationships(transformation_rules, stop_on_exception)

    # Step 2: Generation relationships

    query_statement_template_by_reference = """
    SELECT ?source ?target
    WHERE {
        ?source a prefix:source_class .
        ?target a prefix:target_class .
        ?source prefix:property_ ?target
    }"""

    query_statement_template_by_value = """
    SELECT ?source_id ?target_id
    WHERE {
        ?source a prefix:source_class .
        ?source prefix:property_ ?target .
        ?source prefix:source_ext_id_prop_name ?source_id .
        ?target a prefix:target_class .
        ?target prefix:target_ext_id_prop_name ?target_id .
    }
    """

    relationship_data_frames = []
    for id_, definition in relationship_definitions.relationships.items():
        try:
            logging.debug("Processing relationship: " + id_)
            external_id_prop_name = definition.relationship_external_id_rule
            if external_id_prop_name:
                query = (
                    query_statement_template_by_value.replace("prefix", relationship_definitions.prefix)
                    .replace("source_ext_id_prop_name", external_id_prop_name)
                    .replace("target_ext_id_prop_name", external_id_prop_name)
                    .replace("source_class", definition.source_class)
                    .replace("target_class", definition.target_class)
                    .replace("property_", definition.property_)
                )
            else:
                query = (
                    query_statement_template_by_reference.replace("prefix", relationship_definitions.prefix)
                    .replace("source_class", definition.source_class)
                    .replace("target_class", definition.target_class)
                    .replace("property_", definition.property_)
                )

            logging.debug("Rel query: " + query)
            relationship_data_frame = pd.DataFrame(list(graph_store.query(query)))
            relationship_data_frame.rename(columns={0: "source_external_id", 1: "target_external_id"}, inplace=True)

            # removes namespace
            relationship_data_frame = relationship_data_frame.applymap(remove_namespace)

            # adding prefix
            if transformation_rules.metadata.externalIdPrefix:
                relationship_data_frame["source_external_id"] = (
                    transformation_rules.metadata.externalIdPrefix + relationship_data_frame["source_external_id"]
                )
                relationship_data_frame["target_external_id"] = (
                    transformation_rules.metadata.externalIdPrefix + relationship_data_frame["target_external_id"]
                )

            relationship_data_frame["target_type"] = definition.target_type
            relationship_data_frame["source_type"] = definition.source_type

            # to make sure that by default we set Relationship to active, i.e. non-historic)
            relationship_data_frame["labels"] = [definition.labels] * len(relationship_data_frame)

            # set default external id
            relationship_data_frame["external_id"] = (
                relationship_data_frame["source_external_id"] + ":" + relationship_data_frame["target_external_id"]
            )
            relationship_data_frame["data_set_id"] = transformation_rules.metadata.data_set_id
            relationship_data_frames += [relationship_data_frame]
        except Exception as e:
            logging.error("Error processing relationship: " + id_)
            # traceback.print_exc()
            if stop_on_exception:
                raise e
            continue

    if relationship_data_frames:
        relationship_data_frames = pd.concat(relationship_data_frames)
        relationship_data_frames.reset_index(inplace=True, drop=True)

        # Remove duplicate rows, if any. This should not happen, but it is better to be safe than sorry
        relationship_data_frames.drop_duplicates(subset=["external_id"], inplace=True)

        # Remove duplicate rows, if any. This should not happen, but it is better to be safe than sorry
        relationship_data_frames.drop_duplicates(subset=["external_id"], inplace=True)
        return relationship_data_frames
    else:
        return pd.DataFrame(
            columns=[
                "source_external_id",
                "target_external_id",
                "target_type",
                "source_type",
                "labels",
                "external_id",
                "data_set_id",
            ]
        )


def rdf2relationship_data_frame(
    graph_store: NeatGraphStore,
    transformation_rules: TransformationRules,
    stop_on_exception: bool = False,
) -> pd.DataFrame:
    warn("'rdf2relationship_data_frame' is deprecated, please use 'rdf2relationships' instead!")
    logging.warning("'rdf2relationship_data_frame' is deprecated, please use 'rdf2relationships' instead!")
    return rdf2relationships(graph_store, transformation_rules, stop_on_exception)


def _filter_relationship_xids(relationship_data_frame: pd.DataFrame, asset_xids: Union[List, Set]) -> Set:
    return set(
        relationship_data_frame[
            (relationship_data_frame["source_external_id"].isin(asset_xids))
            | (relationship_data_frame["target_external_id"].isin(asset_xids))
        ]["external_id"]
    )


def _categorize_rdf_relationship_xids(rdf_relationships: pd.DataFrame, categorized_asset_ids: dict) -> Dict[str, set]:
    """Categorizes the external ids of the RDF relationship."""

    missing_asset_ids = (
        set(rdf_relationships.target_external_id)
        .union(rdf_relationships.source_external_id)
        .difference(categorized_asset_ids["historic"].union(categorized_asset_ids["non-historic"]))
    )

    if missing_asset_ids:
        msg = f"Relationships are referring to these assets {missing_asset_ids}, which are missing in CDF."
        msg += "Relationships will not be created for assets that are missing in CDF."
        msg += "Please make sure that all assets are present in CDF before creating relationships."
        logging.warning(msg)

    # First mask all relationships which contain assets that do not exist in CDF
    mask_impossible = _filter_relationship_xids(rdf_relationships, missing_asset_ids)

    # Then mask all relationships which contain assets that are historic while masking
    # all impossible relationships
    mask_historic = _filter_relationship_xids(rdf_relationships, categorized_asset_ids["historic"]).difference(
        mask_impossible
    )

    mask_non_historic = (
        _filter_relationship_xids(rdf_relationships, categorized_asset_ids["non-historic"])
        .difference(mask_historic)
        .difference(mask_impossible)
    )

    return {"impossible": mask_impossible, "historic": mask_historic, "non-historic": mask_non_historic}


def _get_label_based_cdf_relationship_xids(client, data_set_id, labels, partitions) -> set:
    """Get external ids of relationships in CDF for a given data set filtered on labels"""

    labels = LabelFilter(contains_any=labels) if labels is not None else None
    relationship_data_frame = client.relationships.list(
        data_set_ids=data_set_id, limit=-1, labels=labels, partitions=partitions
    ).to_pandas()
    return set() if relationship_data_frame.empty else set(relationship_data_frame.external_id)


def _categorize_cdf_relationship_xids(client, data_set_id, partitions) -> dict[str, set]:
    return {
        "historic": _get_label_based_cdf_relationship_xids(client, data_set_id, ["historic"], partitions),
        "non-historic": _get_label_based_cdf_relationship_xids(client, data_set_id, ["non-historic"], partitions),
    }


def _relationship_to_create(relationships: pd.DataFrame) -> List[Relationship]:
    start_time = datetime_utc_now()
    if relationships.empty:
        return []
    logging.info("Wrangling assets to be created into their final form")
    relationship_list = [Relationship(**row) for row in relationships.to_dict(orient="records")]
    logging.info(f"Wrangling completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return relationship_list


def _relationships_to_decommission(external_ids: list[str]) -> list[RelationshipUpdate]:
    start_time = datetime_utc_now()
    relationships = []
    if not external_ids:
        return []

    logging.info("Wrangling relationships to be decommissioned into their final form")

    for external_id in external_ids:
        # Create relationship update object instance
        relationship = RelationshipUpdate(external_id=external_id)

        # Remove "non-historic" label and add "historic" label
        relationship.labels.remove("non-historic")
        relationship.labels.add(["historic"])

        # Set end time of relationships
        relationship.end_time.set(epoch_now_ms())

        # Add relationship to list of relationship updates
        relationships += [relationship]

    logging.info(f"Wrangling of {len(relationships)} completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return relationships


def _relationships_to_resurrect(external_ids: list[str]) -> list[RelationshipUpdate]:
    start_time = datetime_utc_now()
    relationships = []
    if not external_ids:
        return []

    logging.info("Wrangling relationships to be resurrected into their final form")

    for external_id in external_ids:
        # Create relationship update object instance
        relationship = RelationshipUpdate(external_id=external_id)

        # Remove "non-historic" label and add "historic" label
        relationship.labels.remove("historic")
        relationship.labels.add(["non-historic"])

        # Set end time of relationships
        relationship.end_time.set(None)

        # Add relationship to list of relationship updates
        relationships += [relationship]

    logging.info(f"Wrangling of {len(relationships)} completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return relationships


def categorize_relationships(
    client: CogniteClient,
    rdf_relationships: pd.DataFrame,
    data_set_id: int,
    partitions: int = 40,
    return_report: bool = False,
) -> Union[
    tuple[dict[str, list[Union[Relationship, RelationshipUpdate]]], dict[str, set]],
    dict[str, list[Union[Relationship, RelationshipUpdate]]],
]:
    """Categorize relationships on those that are to be created, decommissioned or resurrected"""
    # TODO also figure out which relationships to be deleted

    _, categorized_asset_ids = _categorize_cdf_assets(client, data_set_id=data_set_id, partitions=partitions)
    categorized_rdf_relationships = _categorize_rdf_relationship_xids(rdf_relationships, categorized_asset_ids)
    categorized_cdf_relationships = _categorize_cdf_relationship_xids(client, data_set_id, partitions=partitions)

    cdf_relationships_all = categorized_cdf_relationships["historic"].union(
        categorized_cdf_relationships["non-historic"]
    )
    rdf_relationships_all = categorized_rdf_relationships["historic"].union(
        categorized_rdf_relationships["non-historic"]
    )

    # relationships to create
    # NonHistoric_rdf - (Historic_cdf ∪ Non-historic_cdf)
    create_xids = categorized_rdf_relationships["non-historic"].difference(cdf_relationships_all)

    # relationships to decommission
    # rdf: Historic_rdf ∩ NonHistoric_cdf ∪ (All_cdf - All_rdf)
    decommission_xids = (
        categorized_rdf_relationships["historic"]
        .intersection(categorized_cdf_relationships["non-historic"])
        .union(categorized_cdf_relationships["non-historic"].difference(rdf_relationships_all))
    )

    # relationships to resurrect
    # NonHistoric_rdf ∩ Historic_cdf
    resurrect_xids = categorized_rdf_relationships["non-historic"].intersection(
        categorized_cdf_relationships["historic"]
    )

    logging.info(f"Number of relationships to create: { len(create_xids)}")
    logging.info(f"Number of relationships to decommission: { len(decommission_xids)}")
    logging.info(f"Number of relationships to resurrect: { len(resurrect_xids)}")

    report = {"create": create_xids, "resurrect": resurrect_xids, "decommission": decommission_xids}
    categorized_relationships = {
        "create": _relationship_to_create(rdf_relationships[rdf_relationships.external_id.isin(create_xids)]),
        "resurrect": _relationships_to_resurrect(resurrect_xids),
        "decommission": _relationships_to_decommission(decommission_xids),
    }

    return (categorized_relationships, report) if return_report else categorized_relationships


def _micro_batch_push(
    client: CogniteClient,
    relationships: list,
    batch_size: int = 1000,
    push_type: str = "update",
    message: str = "Updated",
    max_retries: int = 1,
    retry_delay: int = 5,
):
    """Updates assets in batches of 1000

    Parameters
    ----------
    client : CogniteClient
        Instance of CogniteClient
    relationships : list
        List of relationships to be created or updated
    batch_size : int, optional
        Size of batch, by default 1000
    push_type : str, optional
        Type of push, either "update" or "create", by default "update"
    message : str, optional
        Message to logged, by default "Updated"
    """
    total = len(relationships)
    counter = 0
    if push_type not in ["update", "create"]:
        logging.info(f"push_type {push_type} not supported")
        raise ValueError(f"push_type {push_type} not supported")

    for batch in chunker(relationships, batch_size):
        counter += len(batch)
        start_time = datetime_utc_now()

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="microbatch-relationships")
        def update_relationships(batch):
            if push_type == "update":
                client.relationships.update(batch)
            elif push_type == "create":
                client.relationships.create(batch)

        try:
            update_relationships(batch)
        except CogniteDuplicatedError as e:
            # This situation should not happen but if it does, we need to handle it
            exists = {d["externalId"] for d in e.duplicated}
            missing_relationships = [t for t in batch if t.external_id not in exists]
            client.relationships.create(missing_relationships)

        delta_time = (datetime_utc_now() - start_time).seconds

        msg = f"{message} {counter} of {total} relationships, batch processing time: {delta_time:.2f} "
        msg += f"seconds ETC: {delta_time * (total - counter) / (60*batch_size) :.2f} minutes"
        logging.info(msg)


def upload_relationships(
    client: CogniteClient,
    categorized_relationships: Dict[str, list[Union[Relationship, RelationshipUpdate]]],
    batch_size: int = 5000,
    max_retries: int = 1,
    retry_delay: int = 3,
):
    """Uploads categorized relationships to CDF

    Parameters
    ----------
    client : CogniteClient
        Instance of CogniteClient
    categorized_relationships : Dict[str, list[Union[Relationship, RelationshipUpdate]]]
        Categories of relationships to be uploaded
    batch_size : int, optional
        Size of batch, by default 5000
    """
    if batch_size:
        logging.info(f"Uploading relationships in batches of {batch_size}")
        if categorized_relationships["create"]:
            _micro_batch_push(
                client,
                categorized_relationships["create"],
                batch_size,
                push_type="create",
                message="Created",
                max_retries=max_retries,
                retry_delay=retry_delay,
            )

        if categorized_relationships["resurrect"]:
            _micro_batch_push(
                client,
                categorized_relationships["resurrect"],
                batch_size,
                message="Resurrected",
                max_retries=max_retries,
                retry_delay=retry_delay,
            )

        if categorized_relationships["decommission"]:
            _micro_batch_push(
                client,
                categorized_relationships["decommission"],
                batch_size,
                message="Decommissioned",
                max_retries=max_retries,
                retry_delay=retry_delay,
            )

    else:
        logging.info("Batch size not set, pushing all relationships to CDF in one go!")

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="create-relationships")
        def create_relationships():
            if categorized_relationships["create"]:
                client.relationships.create(categorized_relationships["create"])

            if categorized_relationships["resurrect"]:
                client.relationships.update(categorized_relationships["resurrect"])

            if categorized_relationships["decommission"]:
                client.relationships.update(categorized_relationships["decommission"])

        try:
            create_relationships()
        except CogniteDuplicatedError as e:
            # This situation should not happen but if it does, the code attempts to handle it
            exists = {d["externalId"] for d in e.duplicated}
            missing_relationships = [t for t in categorized_relationships["create"] if t.external_id not in exists]
            client.relationships.create(missing_relationships)



"""This module performs importing of various formats to one of serializations for which
there are loaders to TransformationRules pydantic class."""

# TODO: if this module grows too big, split it into several files and place under ./converter directory

import warnings
from pathlib import Path

import numpy as np
import pandas as pd
from rdflib import DC, DCTERMS, OWL, RDF, RDFS, Graph

from cognite.neat.rules import parse_rules_from_excel_file
from cognite.neat.rules import _exceptions
from cognite.neat.utils.utils import generate_exception_report, get_namespace, remove_namespace


def _create_default_metadata_parsing_config() -> dict[str, tuple[str, ...]]:
    # TODO: these are to be read from Metadata pydantic model
    return {
        "header": (
            "namespace",
            "prefix",
            "dataModelName",
            "cdfSpaceName",
            "version",
            "isCurrentVersion",
            "created",
            "updated",
            "title",
            "description",
            "creator",
            "contributor",
            "rights",
            "license",
        )
    }


def _create_default_classes_parsing_config() -> dict[str, tuple[str, ...]]:
    # TODO: these are to be read from Class pydantic model
    return {
        "helper_row": ("Data Model Definition", "", "", "", "State", "", "", "Knowledge acquisition log", "", "", ""),
        "header": (
            "Class",
            "Name",
            "Description",
            "Parent Class",
            "Deprecated",
            "Deprecation Date",
            "Replaced By",
            "Source",
            "Source Entity Name",
            "Match",
            "Comment",
        ),
    }


def _create_default_properties_parsing_config() -> dict[str, tuple[str, ...]]:
    # TODO: these are to be read from Property pydantic model
    return {
        "helper_row": (
            "Data Model Definition",
            "",
            "",
            "",
            "",
            "",
            "",
            "State",
            "",
            "",
            "Knowledge acquisition log",
            "",
            "",
            "",
        ),
        "header": (
            "Class",
            "Property",
            "Name",
            "Description",
            "Type",
            "Min Count",
            "Max Count",
            "Deprecated",
            "Deprecation Date",
            "Replaced By",
            "Source",
            "Source Entity Name",
            "Match",
            "Comment",
        ),
    }


def owl2excel(owl_filepath: Path, excel_filepath: Path = None, validate_results: bool = True):
    """Convert owl ontology to transformation rules serialized as Excel file.

    Parameters
    ----------
    owl_filepath : Path
        Path to OWL ontology
    excel_filepath : Path
        Path to save transformation rules, defaults to None
    validate_results : bool, optional
        Whether to validate generated Excel file and create validation report, by default True
    """

    owl_filepath = Path(owl_filepath)
    if excel_filepath:
        excel_filepath = Path(excel_filepath)
    else:
        excel_filepath = owl_filepath.parent / "transformation_rules.xlsx"

    graph = Graph()
    try:
        graph.parse(owl_filepath)
    except Exception as e:
        raise Exception(f"Could not parse owl file: {e}") from e

    # bind key namespaces
    graph.bind("owl", OWL)
    graph.bind("rdf", RDF)
    graph.bind("rdfs", RDFS)
    graph.bind("dcterms", DCTERMS)
    graph.bind("dc", DC)

    with pd.ExcelWriter(excel_filepath) as writer:
        _parse_owl_metadata_df(graph).to_excel(writer, sheet_name="Metadata", header=False)
        _parse_owl_classes_df(graph).to_excel(writer, sheet_name="Classes", index=False, header=False)
        _parse_owl_properties_df(graph).to_excel(writer, sheet_name="Properties", index=False, header=False)

    if validate_results:
        _validate_excel_file(excel_filepath)


def _parse_owl_metadata_df(graph: Graph, parsing_config: dict = None) -> pd.DataFrame:
    """Parse owl metadata from graph to pandas dataframe.

    Parameters
    ----------
    graph : Graph
        Graph containing owl metadata
    parsing_config : dict, optional
        Configuration for parsing, by default None

    Returns
    -------
    pd.DataFrame
        Dataframe containing owl metadata
    """
    if parsing_config is None:
        parsing_config = _create_default_metadata_parsing_config()

    query = """
    SELECT ?namespace ?prefix ?dataModelName ?cdfSpaceName ?version ?isCurrentVersion
           ?created ?updated ?title ?description ?creator ?contributor ?rights ?license
    WHERE {
        ?namespace a owl:Ontology .
        OPTIONAL {?namespace owl:versionInfo ?version }.
        OPTIONAL {?namespace dcterms:creator ?creator }.
        OPTIONAL {?namespace dcterms:title ?title }.
        OPTIONAL {?namespace dcterms:contributor ?contributor }.
        OPTIONAL {?namespace dcterms:modified ?updated }.
        OPTIONAL {?namespace dcterms:created ?created }.
        OPTIONAL {?namespace dcterms:description ?description }.
        OPTIONAL {?namespace dcterms:rights|dc:rights ?rights }.
        OPTIONAL {?namespace dcterms:license|dc:license ?license }.
        FILTER (!isBlank(?namespace))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
        FILTER (!bound(?title) || LANG(?title) = "" || LANGMATCHES(LANG(?title), "en"))
    }
    """

    raw_df = pd.DataFrame(list(graph.query(query)), columns=parsing_config["header"])
    if raw_df.empty:
        return raw_df
    raw_df.replace(np.nan, "", regex=True, inplace=True)

    clean_list = [
        [
            raw_df.namespace.unique()[0],
            raw_df.prefix.unique()[0],
            raw_df.dataModelName.unique()[0],
            raw_df.cdfSpaceName.unique()[0],
            raw_df.version.unique()[0],
            True,
            raw_df.created.unique()[0],
            raw_df.updated.unique()[0],
            raw_df.title.unique()[0],
            raw_df.description.unique()[0],
            ", ".join(list(raw_df.creator.unique())),
            ", ".join(list(raw_df.contributor.unique())),
            raw_df.rights.unique()[0],
            raw_df.license.unique()[0],
        ]
    ]

    return pd.DataFrame(clean_list, columns=parsing_config["header"]).T


def _parse_owl_classes_df(graph: Graph, parsing_config: dict = None) -> pd.DataFrame:
    """Get all classes from the graph and their parent classes.

    Parameters
    ----------
    graph : Graph
        Graph to query
    parsing_config : dict, optional
        Configuration for parsing the dataframe, by default None

    Returns
    -------
    pd.DataFrame
        Dataframe with columns: class, parentClass
    """
    if parsing_config is None:
        parsing_config = _create_default_classes_parsing_config()

    query = """
SELECT ?class ?name ?description ?parentClass ?deprecated ?deprecationDate
?replacedBy ?source ?sourceEntity ?match ?comment
    WHERE {
        ?class a owl:Class .
        OPTIONAL {?class rdfs:subClassOf ?parentClass }.
        OPTIONAL {?class rdfs:label ?name }.
        OPTIONAL {?class rdfs:comment ?description} .
        OPTIONAL {?class owl:deprecated ?deprecated} .
        FILTER (!isBlank(?class))
        FILTER (!bound(?parentClass) || !isBlank(?parentClass))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
    }
    """

    raw_df = pd.DataFrame(list(graph.query(query)), columns=parsing_config["header"])
    if raw_df.empty:
        raw_df = pd.concat([raw_df, pd.DataFrame([len(raw_df) * [""]])], ignore_index=True)
        return raw_df
    raw_df.replace(np.nan, "", regex=True, inplace=True)

    raw_df.Source = raw_df.Class.apply(lambda x: get_namespace(x))
    raw_df.Class = raw_df.Class.apply(lambda x: remove_namespace(x))
    raw_df["Source Entity Name"] = raw_df.Class
    raw_df["Match"] = len(raw_df) * ["exact"]
    raw_df["Parent Class"] = raw_df["Parent Class"].apply(lambda x: remove_namespace(x))

    grouped_df = raw_df.groupby("Class")

    clean_list = [
        [
            class_,
            group_df.Name.unique()[0],
            "\n".join(list(group_df.Description.unique())),
            ", ".join(list(group_df["Parent Class"].unique())),
            group_df.Deprecated.unique()[0],
            group_df["Deprecation Date"].unique()[0],
            group_df["Replaced By"].unique()[0],
            group_df["Source"].unique()[0],
            group_df["Source Entity Name"].unique()[0],
            group_df["Match"].unique()[0],
            group_df["Comment"].unique()[0],
        ]
        for class_, group_df in grouped_df
    ]
    return pd.DataFrame([parsing_config["helper_row"], parsing_config["header"]] + clean_list)


def _parse_owl_properties_df(graph: Graph, parsing_config: dict = None) -> pd.DataFrame:
    """Get all properties from the OWL ontology

    Parameters
    ----------
    graph : Graph
        Graph to query
    parsing_config : dict, optional
        Configuration for parsing the dataframe, by default None

    Returns
    -------
    pd.DataFrame
        Dataframe with columns: class, property, name, ...
    """
    if parsing_config is None:
        parsing_config = _create_default_properties_parsing_config()

    query = """
    SELECT ?class ?property ?name ?description ?type ?minCount ?maxCount
    ?deprecated ?deprecationDate ?replacedBy ?source ?sourceEntity ?match ?comment
    WHERE {
        ?property a ?propertyType.
        FILTER (?propertyType IN (owl:ObjectProperty, owl:DatatypeProperty ) )
        OPTIONAL {?property rdfs:domain ?class }.
        OPTIONAL {?property rdfs:range ?type }.
        OPTIONAL {?property rdfs:label ?name }.
        OPTIONAL {?property rdfs:comment ?description} .
        FILTER (!isBlank(?property))
        FILTER (!bound(?class) || !isBlank(?class))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
        OPTIONAL {?property owl:deprecated ?deprecated} .
    }
    """

    raw_df = pd.DataFrame(list(graph.query(query)), columns=parsing_config["header"])
    if raw_df.empty:
        return raw_df
    raw_df.replace(np.nan, "", regex=True, inplace=True)

    raw_df.Source = raw_df.Property.apply(lambda x: get_namespace(x))
    raw_df.Class = raw_df.Class.apply(lambda x: remove_namespace(x))
    raw_df.Property = raw_df.Property.apply(lambda x: remove_namespace(x))
    raw_df.Type = raw_df.Type.apply(lambda x: remove_namespace(x))
    raw_df["Source Entity Name"] = raw_df.Property
    raw_df["Match"] = len(raw_df) * ["exact"]

    class_grouped_dfs = raw_df.groupby("Class")

    clean_list = []

    for class_, class_grouped_df in class_grouped_dfs:
        property_grouped_dfs = class_grouped_df.groupby("Property")
        for property_, property_grouped_df in property_grouped_dfs:
            clean_list += [
                [
                    class_,
                    property_,
                    property_grouped_df.Name.unique()[0],
                    "\n".join(list(property_grouped_df.Description.unique())),
                    property_grouped_df.Type.unique()[0],
                    property_grouped_df["Min Count"].unique()[0],
                    property_grouped_df["Max Count"].unique()[0],
                    property_grouped_df.Deprecated.unique()[0],
                    property_grouped_df["Deprecation Date"].unique()[0],
                    property_grouped_df["Replaced By"].unique()[0],
                    property_grouped_df["Source"].unique()[0],
                    property_grouped_df["Source Entity Name"].unique()[0],
                    property_grouped_df["Match"].unique()[0],
                    property_grouped_df["Comment"].unique()[0],
                ]
            ]

    return pd.DataFrame([parsing_config["helper_row"], parsing_config["header"]] + clean_list)


def _validate_excel_file(excel_filepath: Path):
    _, validation_errors, validation_warnings = parse_rules_from_excel_file(excel_filepath, return_report=True)

    report = ""
    if validation_errors:
        warnings.warn(
            _exceptions.Warning1().message,
            category=_exceptions.Warning1,
            stacklevel=2,
        )
        report = generate_exception_report(validation_errors, "Errors")

    if validation_warnings:
        warnings.warn(
            _exceptions.Warning2().message,
            category=_exceptions.Warning2,
            stacklevel=2,
        )
        report += generate_exception_report(validation_warnings, "Warnings")

    if report:
        with open(excel_filepath.parent / "report.txt", "w") as f:
            f.write(report)



from rdflib.query import Result


def rdf_result_to_api_response(result: Result):
    response = {"fields": [], "rows": []}
    if result.vars is None:
        result.vars = ["s", "p", "o"]
    for field in result.vars:
        response["fields"].append(field)
    for row in result:
        rrow = {field: row[i] for i, field in enumerate(result.vars)}
        response["rows"].append(rrow)
    return response



import logging

from pydantic import BaseModel, field_validator, model_validator


class QueryRequest(BaseModel):
    graph_name: str
    workflow_name: str = "default"
    query: str


class RuleRequest(BaseModel):
    graph_name: str
    workflow_name: str = "default"
    rule_type: str
    rule: str


class UploadToCdfRequest(BaseModel):
    file_name: str = ""
    file_type: str = "workflow"
    comments: str = ""
    author: str = ""
    tag: str = ""


class DownloadFromCdfRequest(BaseModel):
    file_name: str = ""
    file_type: str = "workflow"
    version: str = ""


class LoadGraphRequest(BaseModel):
    graph_source_template_name: str
    source_location: str


class RunWorkflowRequest(BaseModel):
    name: str
    config: dict
    start_step: str
    sync: bool = False

    @field_validator("config", mode="before")
    def empty_string_to_dict(cls, value):
        if value == "":
            return {}
        return value

    @model_validator(mode="before")
    def log_call(cls, values):
        logging.info(f"RunWorkflowRequest: {values}")
        return values


class NodesAndEdgesRequest(BaseModel):
    graph_name: str
    workflow_name: str = "default"
    node_class_filter: list[str]
    src_edge_filter: list[str]
    dst_edge_filter: list[str]
    cache: bool = False
    limit: int = 1000
    node_name_property: str = ""
    sparql_query: str = ""



import logging
import os
from pathlib import Path
from fastapi import APIRouter
from cognite.neat.app.api.configuration import Config, neat_app


router = APIRouter()


@router.get("/api/configs/global")
def get_configs():
    return neat_app.config.dict()


@router.post("/api/configs/global")
def set_configs(request: Config):
    logging.info(f"Updating global config: {request}")
    global config
    global cache_store
    global neat_app
    config = request
    config.to_yaml(Path(os.environ.get("NEAT_CONFIG_PATH", "config.yaml")))
    cache_store = {}
    neat_app.stop()
    neat_app.start(config=config)
    return config



import json
import logging
import os
from enum import StrEnum
from pathlib import Path

from typing import Literal, Optional, Self

import yaml
from pydantic import BaseModel, Field
from yaml import safe_load

from cognite.neat.utils.cdf import InteractiveCogniteClient, ServiceCogniteClient

LOG_FORMAT = "%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s"
LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


def configure_logging(level: str = "DEBUG", log_format: str = LOG_FORMAT):
    """Configure logging based on config."""
    logging.basicConfig(format=log_format, level=logging.getLevelName(level), datefmt=LOG_DATE_FORMAT)


class RulesStoreType(StrEnum):
    """Rules Store type"""

    CDF = "cdf"
    FILE = "file"
    URL = "url"
    GOOGLE_SHEET = "google_sheet"


class WorkflowsStoreType(StrEnum):
    """Workflows Store type"""

    CDF = "cdf"
    FILE = "file"
    URL = "url"


class Config(BaseModel):
    workflows_store_type: RulesStoreType = WorkflowsStoreType.FILE
    data_store_path: Path = Field(default_factory=lambda: Path.cwd() / "data")

    workflow_downloader_filter: Optional[list[str]] = Field(
        description="List of workflow names+tags to filter on when downloading workflows from CDF. "
        "Example name:workflow_name=version,tag:tag_name",
        default=None,
    )

    cdf_client: InteractiveCogniteClient | ServiceCogniteClient = ServiceCogniteClient()
    cdf_default_dataset_id: int = 0
    load_examples: bool = True

    log_level: Literal["ERROR", "WARNING", "INFO", "DEBUG"] = "INFO"
    log_format: str = LOG_FORMAT
    download_workflows_from_cdf: bool = Field(
        default=False, description="Downloads all workflows from CDF automatically and stores them locally"
    )

    stop_on_error: bool = False

    @property
    def rules_store_path(self) -> Path:
        return self.data_store_path / "rules"

    @property
    def workflows_store_path(self) -> Path:
        return self.data_store_path / "workflows"

    @property
    def source_graph_path(self) -> Path:
        return self.data_store_path / "source-graphs"

    @classmethod
    def from_yaml(cls, filepath: Path) -> Self:
        return cls(**safe_load(filepath.read_text()))

    def to_yaml(self, filepath: Path):
        # Parse as json to avoid Path and Enum objects
        dump = json.loads(self.json())

        with filepath.open("w") as f:
            yaml.safe_dump(dump, f)

    @classmethod
    def from_env(cls) -> Self:
        cdf_config = ServiceCogniteClient(
            project=os.environ.get("NEAT_CDF_PROJECT"),
            client_name=os.environ.get("NEAT_CDF_CLIENT_NAME"),
            client_id=os.environ.get("NEAT_CDF_CLIENT_ID"),
            client_secret=os.environ.get("NEAT_CDF_CLIENT_SECRET"),
            base_url=os.environ.get("NEAT_CDF_BASE_URL"),
            token_url=os.environ.get("NEAT_CDF_TOKEN_URL"),
            scopes=[os.environ.get("NEAT_CDF_SCOPES")],
            timeout=int(os.environ.get("NEAT_CDF_CLIENT_TIMEOUT", "60")),
            max_workers=int(os.environ.get("NEAT_CDF_CLIENT_MAX_WORKERS", "3")),
        )

        if workflow_downloader_filter := os.environ.get("NEAT_WORKFLOW_DOWNLOADER_FILTER", None):
            workflow_downloader_filter = workflow_downloader_filter.split(",")

        return cls(
            cdf_client=cdf_config,
            workflows_store_type=os.environ.get("NEAT_WORKFLOWS_STORE_TYPE", WorkflowsStoreType.FILE),
            workflows_store_path=os.environ.get("NEAT_DATA_PATH", "/app/data"),
            cdf_default_dataset_id=os.environ.get("NEAT_CDF_DEFAULT_DATASET_ID", 6476640149881990),
            log_level=os.environ.get("NEAT_LOG_LEVEL", "INFO"),
            workflow_downloader_filter=workflow_downloader_filter,
            load_examples=bool(os.environ.get("NEAT_LOAD_EXAMPLES", True)),
        )



import logging
import time
from cognite.neat.graph.loaders import upload_labels
from cognite.neat.graph.loaders.core.rdf_to_assets import (
    NeatMetadataKeys,
    categorize_assets,
    rdf2assets,
    remove_non_existing_labels,
    unique_asset_labels,
    upload_assets,
)
from cognite.neat.graph.loaders.core.rdf_to_relationships import (
    categorize_relationships,
    rdf2relationships,
    upload_relationships,
)
from cognite.neat.graph.loaders.validator import validate_asset_hierarchy
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.steps.step_model import Step
from cognite.client.data_classes import AssetFilter

from cognite.client import CogniteClient
from ..data_contracts import CategorizedAssets, CategorizedRelationships, RulesData, SolutionGraph

__all__ = [
    "CreateCDFLabels",
    "GenerateCDFAssetsFromGraph",
    "GenerateCDFRelationshipsFromGraph",
    "UploadCDFAssets",
    "UploadCDFRelationships",
]


class CreateCDFLabels(Step):
    description = "The step creates default NEAT labels in CDF"
    category = "cdf_resources"

    def run(self, rules: RulesData, cdf_client: CogniteClient) -> None:
        upload_labels(cdf_client, rules.rules, extra_labels=["non-historic", "historic"])


class GenerateCDFAssetsFromGraph(Step):
    description = (
        "The step generates assets from the graph ,categorizes them and stores them in CategorizedAssets object"
    )
    category = "cdf_resources"

    def run(
        self, rules: RulesData, cdf_client: CogniteClient, solution_graph: SolutionGraph
    ) -> (FlowMessage, CategorizedAssets):
        meta_keys = NeatMetadataKeys.load(
            self.configs.get_config_group_values_by_name("cdf.asset.metadata.", remove_group_prefix=True)
        )
        prom_cdf_resource_stats = self.metrics.register_metric(
            "cdf_resources_stats",
            "CDF resource stats before and after running the workflow",
            m_type="gauge",
            metric_labels=["resource_type", "state"],
        )
        prom_data_issues_stats = self.metrics.register_metric(
            "data_issues_stats", "Data validation issues", m_type="gauge", metric_labels=["resource_type"]
        )

        rdf_asset_dicts = rdf2assets(
            solution_graph.graph,
            rules.rules,
            stop_on_exception=True,
            meta_keys=meta_keys,
        )
        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_before = cdf_client.assets.aggregate(filter=AssetFilter(data_set_ids=[{"id": rules.dataset_id}]))[
            0
        ]["count"]

        # Label Validation
        labels_before = unique_asset_labels(rdf_asset_dicts.values())
        logging.info(f"Assets have {len(labels_before)} unique labels: {', '.join(sorted(labels_before))}")

        rdf_asset_dicts = remove_non_existing_labels(cdf_client, rdf_asset_dicts)

        labels_after = unique_asset_labels(rdf_asset_dicts.values())
        removed_labels = labels_before - labels_after
        logging.info(
            f"Removed {len(removed_labels)} labels as these do not exists in CDF. "
            f"Removed labels: {', '.join(sorted(removed_labels))}"
        )
        ######################

        prom_cdf_resource_stats.labels(resource_type="asset", state="count_before_neat_update").set(total_assets_before)
        logging.info(f"Total count of assets in CDF before upload: { total_assets_before }")

        orphan_assets, circular_assets = validate_asset_hierarchy(rdf_asset_dicts)

        prom_data_issues_stats.labels(resource_type="circular_assets").set(len(circular_assets))
        prom_data_issues_stats.labels(resource_type="orphan_assets").set(len(orphan_assets))

        if orphan_assets:
            logging.error(f"Found orphaned assets: {', '.join(orphan_assets)}")

            orphanage_asset_external_id = (
                f"{rules.rules.metadata.externalIdPrefix}orphanage-{rules.dataset_id}"
                if rules.rules.metadata.externalIdPrefix
                else "orphanage"
            )

            # Kill the process if you dont have orphanage asset in your asset hierarchy
            # and inform the user that it is missing !
            if orphanage_asset_external_id not in rdf_asset_dicts:
                msg = f"You dont have Orphanage asset {orphanage_asset_external_id} in asset hierarchy!"
                logging.error(msg)
                raise Exception(msg)

            logging.error("Orphaned assets will be assigned to 'Orphanage' root asset")

            for external_id in orphan_assets:
                rdf_asset_dicts[external_id]["parent_external_id"] = orphanage_asset_external_id

            orphan_assets, circular_assets = validate_asset_hierarchy(rdf_asset_dicts)

            logging.info(orphan_assets)
        else:
            logging.info("No orphaned assets found, your assets look healthy !")

        if circular_assets:
            msg = f"Found circular dependencies: {str(circular_assets)}"
            logging.error(msg)
            raise Exception(msg)
        elif orphan_assets:
            msg = f"Not able to fix orphans: {', '.join(orphan_assets)}"
            logging.error(msg)
            raise Exception(msg)
        else:
            logging.info("No circular dependency among assets found, your assets hierarchy look healthy !")

        categorized_assets, report = categorize_assets(
            cdf_client, rdf_asset_dicts, rules.dataset_id, return_report=True
        )

        count_create_assets = len(categorized_assets["create"])
        count_update_assets = len(categorized_assets["update"])
        count_decommission_assets = len(categorized_assets["decommission"])
        count_resurrect_assets = len(categorized_assets["resurrect"])

        prom_cdf_resource_stats.labels(resource_type="asset", state="create").set(count_create_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="update").set(count_update_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="decommission").set(count_decommission_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="resurrect").set(count_resurrect_assets)

        logging.info(f"Total count of assets to be created: { count_create_assets }")
        logging.info(f"Total count of assets to be updated: { count_update_assets }")
        logging.info(f"Total count of assets to be decommission: { count_decommission_assets }")
        logging.info(f"Total count of assets to be resurrect: { count_resurrect_assets }")

        msg = f"Total count of assets { len(rdf_asset_dicts) } of which: { count_create_assets } to be created"
        msg += f", { count_update_assets } to be updated"
        msg += f", { count_decommission_assets } to be decommissioned"
        msg += f", { count_resurrect_assets } to be resurrected"
        number_of_updates = len(report["decommission"])
        logging.info(f"Total number of updates: {number_of_updates}")

        return FlowMessage(output_text=msg), CategorizedAssets(assets=categorized_assets)


class UploadCDFAssets(Step):
    description = "The step uploads categorized assets to CDF"
    category = "cdf_resources"

    def run(
        self, rules: RulesData, cdf_client: CogniteClient, categorized_assets: CategorizedAssets, flow_msg: FlowMessage
    ) -> FlowMessage:
        if flow_msg and flow_msg.payload and "action" in flow_msg.payload:
            if flow_msg.payload["action"] != "approve":
                raise Exception("Update not approved")
        prom_cdf_resource_stats = self.metrics.register_metric(
            "cdf_resources_stats",
            "CDF resource stats before and after running the workflow",
            m_type="gauge",
            metric_labels=["resource_type", "state"],
        )
        upload_assets(cdf_client, categorized_assets.assets, max_retries=2, retry_delay=4)
        count_create_assets = len(categorized_assets.assets["create"])
        for _ in range(1000):
            total_assets_after = cdf_client.assets.aggregate(
                filter=AssetFilter(data_set_ids=[{"id": rules.dataset_id}])
            )[0]["count"]
            if total_assets_after >= count_create_assets:
                break
            logging.info(f"Waiting for assets to be created, current count {total_assets_after}")
            time.sleep(2)

        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_after = cdf_client.assets.aggregate(filter=AssetFilter(data_set_ids=[{"id": rules.dataset_id}]))[
            0
        ]["count"]

        prom_cdf_resource_stats.labels(resource_type="asset", state="count_after_neat_update").set(total_assets_after)
        logging.info(f"Total count of assets in CDF after update: { total_assets_after }")
        categorized_assets.assets = None  # free up memory after upload .
        return FlowMessage(output_text=f"Total count of assets in CDF after update: { total_assets_after }")


class GenerateCDFRelationshipsFromGraph(Step):
    description = "The step generates relationships from the graph and saves them to CategorizedRelationships object"
    category = "cdf_resources"

    def run(
        self, rules: RulesData, cdf_client: CogniteClient, solution_graph: SolutionGraph
    ) -> (FlowMessage, CategorizedRelationships):
        # create, categorize and upload relationships
        rdf_relationships = rdf2relationships(
            solution_graph.graph.get_graph(),
            rules.rules,
        )

        categorized_relationships = categorize_relationships(cdf_client, rdf_relationships, rules.dataset_id)
        count_defined_relationships = len(rdf_relationships)
        count_create_relationships = len(categorized_relationships["create"])
        count_decommission_relationships = len(categorized_relationships["decommission"])
        count_resurrect_relationships = len(categorized_relationships["resurrect"])
        prom_cdf_resource_stats = self.metrics.register_metric(
            "cdf_resources_stats",
            "CDF resource stats before and after running the workflow",
            m_type="gauge",
            metric_labels=["resource_type", "state"],
        )
        prom_cdf_resource_stats.labels(resource_type="relationships", state="defined").set(count_defined_relationships)
        prom_cdf_resource_stats.labels(resource_type="relationships", state="create").set(count_create_relationships)
        prom_cdf_resource_stats.labels(resource_type="relationships", state="decommission").set(
            count_decommission_relationships
        )
        prom_cdf_resource_stats.labels(resource_type="relationships", state="resurrect").set(
            count_resurrect_relationships
        )

        msg = (
            f"Total count of relationships { count_defined_relationships } of which:"
            f" { count_create_relationships } to be created"
        )
        msg += f", { count_decommission_relationships } to be decommissioned"
        msg += f", { count_resurrect_relationships } to be resurrected"

        return FlowMessage(output_text=msg), CategorizedRelationships(relationships=categorized_relationships)


class UploadCDFRelationships(Step):
    description = "The step uploads relationships to CDF"
    category = "cdf_resources"

    def run(self, cdf_client: CogniteClient, categorized_relationships: CategorizedRelationships) -> FlowMessage:
        upload_relationships(cdf_client, categorized_relationships.relationships, max_retries=2, retry_delay=4)
        return FlowMessage(output_text="CDF relationships uploaded successfully")



from enum import StrEnum
from pydantic import BaseModel


class RdfStoreType(StrEnum):
    """RDF Store type"""

    MEMORY = "memory"
    FILE = "file"
    GRAPHDB = "graphdb"
    SPARQL = "sparql"
    OXIGRAPH = "oxigraph"


class RdfStoreConfig(BaseModel):
    type: RdfStoreType
    file_path: str = None
    query_url: str = None
    update_url: str = None
    user: str = None
    password: str = None
    api_root_url: str = "http://localhost:7200"



from typing import ClassVar, Optional, Self
import warnings
from pydantic import BaseModel, ConfigDict, FieldValidationInfo, field_validator
from rdflib import OWL, RDF, RDFS, XSD, DCTERMS, BNode, Graph, Literal, URIRef, Namespace
from rdflib.collection import Collection as GraphCollection

from cognite.neat.rules.models import DATA_TYPE_MAPPING, Class, Property, TransformationRules, Metadata
from cognite.neat.rules.analysis import to_property_dict, to_class_property_pairs
from cognite.neat.rules._validation import are_properties_redefined
from cognite.neat.rules import _exceptions
from cognite.neat.utils.utils import generate_exception_report, remove_namespace


class OntologyModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True, strict=False, extra="allow")


class Ontology(OntologyModel):
    properties: list["OWLProperty"]
    classes: list["OWLClass"]
    shapes: list["SHACLNodeShape"]
    metadata: "OWLMetadata"
    prefixes: dict[str, Namespace]

    @classmethod
    def from_rules(cls, transformation_rules: TransformationRules) -> Self:
        properties_redefined, redefinition_warnings = are_properties_redefined(transformation_rules, return_report=True)
        if properties_redefined:
            raise _exceptions.Error11(report=generate_exception_report(redefinition_warnings))

        return cls(
            properties=[
                OWLProperty.from_list_of_properties(
                    definition,
                    transformation_rules.metadata.namespace,
                )
                for definition in to_property_dict(transformation_rules).values()
            ],
            classes=[
                OWLClass.from_class(
                    definition,
                    transformation_rules.metadata.namespace,
                )
                for definition in transformation_rules.classes.values()
            ],
            shapes=[
                SHACLNodeShape.from_rules(
                    transformation_rules.classes[class_],
                    list(properties.values()),
                    transformation_rules.metadata.namespace,
                )
                for class_, properties in to_class_property_pairs(transformation_rules).items()
            ],
            metadata=OWLMetadata(**transformation_rules.metadata.model_dump()),
            prefixes=transformation_rules.prefixes,
        )

    def as_shacl(self) -> Graph:
        shacl = Graph()
        shacl.bind(self.metadata.prefix, self.metadata.namespace)
        for prefix, namespace in self.prefixes.items():
            shacl.bind(prefix, namespace)

        for shape in self.shapes:
            for triple in shape.triples:
                shacl.add(triple)

        return shacl

    def as_owl(self) -> Graph:
        owl = Graph()
        owl.bind(self.metadata.prefix, self.metadata.namespace)
        for prefix, namespace in self.prefixes.items():
            owl.bind(prefix, namespace)

        owl.add((URIRef(self.metadata.namespace), RDF.type, OWL.Ontology))
        for property_ in self.properties:
            for triple in property_.triples:
                owl.add(triple)

        for class_ in self.classes:
            for triple in class_.triples:
                owl.add(triple)

        for triple in self.metadata.triples:
            owl.add(triple)

        return owl

    @property
    def owl_triples(self) -> list[tuple]:
        return list(self.as_owl())

    @property
    def shacl_triples(self) -> list[tuple]:
        return list(self.as_shacl())

    @property
    def triples(self) -> list[tuple]:
        return self.owl_triples + self.shacl_triples

    @property
    def ontology(self) -> str:
        return self.as_owl().serialize()

    @property
    def constraints(self) -> str:
        return self.as_shacl().serialize()

    @property
    def semantic_data_model(self) -> str:
        return (self.as_owl() + self.as_shacl()).serialize()


class OWLMetadata(Metadata):
    @property
    def triples(self) -> list[tuple]:
        # Mandatory triples originating from Metadata mandatory fields
        triples = [
            (URIRef(self.namespace), DCTERMS.hasVersion, Literal(self.version)),
            (URIRef(self.namespace), OWL.versionInfo, Literal(self.version)),
            (URIRef(self.namespace), RDFS.label, Literal(self.title)),
            (URIRef(self.namespace), DCTERMS.title, Literal(self.title)),
            (URIRef(self.namespace), DCTERMS.created, Literal(self.created, datatype=XSD.dateTime)),
            (URIRef(self.namespace), DCTERMS.description, Literal(self.description)),
        ]
        if isinstance(self.creator, list):
            triples.extend([(URIRef(self.namespace), DCTERMS.creator, Literal(creator)) for creator in self.creator])
        else:
            triples.append((URIRef(self.namespace), DCTERMS.creator, Literal(self.creator)))

        # Optional triples originating from Metadata optional fields
        if self.updated:
            triples.append((URIRef(self.namespace), DCTERMS.modified, Literal(self.updated, datatype=XSD.dateTime)))
        if self.rights:
            triples.append((URIRef(self.namespace), DCTERMS.rights, Literal(self.rights)))

        if self.contributor and isinstance(self.contributor, list):
            triples.extend(
                [
                    (URIRef(self.namespace), DCTERMS.contributor, Literal(contributor))
                    for contributor in self.contributor
                ]
            )
        elif self.contributor:
            triples.append((URIRef(self.namespace), DCTERMS.contributor, Literal(self.contributor)))

        return triples


class OWLClass(OntologyModel):
    id_: URIRef
    type_: URIRef = OWL.Class
    label: Optional[str]
    comment: Optional[str]
    sub_class_of: Optional[URIRef]
    namespace: Namespace

    @classmethod
    def from_class(cls, definition: Class, namespace: Namespace) -> Self:
        return cls(
            id_=namespace[definition.class_id],
            label=definition.class_name,
            comment=definition.description,
            sub_class_of=namespace[definition.parent_class] if definition.parent_class else None,
            namespace=namespace,
        )

    @property
    def type_triples(self) -> list[tuple]:
        return [(self.id_, RDF.type, self.type_)]

    @property
    def label_triples(self) -> list[tuple]:
        return [(self.id_, RDFS.label, Literal(self.label))]

    @property
    def comment_triples(self) -> list[tuple]:
        return [(self.id_, RDFS.comment, Literal(self.comment))]

    @property
    def subclass_triples(self) -> list[tuple]:
        if self.sub_class_of:
            return [(self.id_, RDFS.subClassOf, self.sub_class_of)]
        else:
            return []

    @property
    def triples(self) -> list[tuple]:
        return self.type_triples + self.label_triples + self.comment_triples + self.subclass_triples


class OWLProperty(OntologyModel):
    id_: URIRef
    type_: set[URIRef]
    label: set[str]
    comment: set[str]
    domain: set[URIRef]
    range_: set[URIRef]
    namespace: Namespace

    @staticmethod
    def same_property_id(definitions: list[Property]) -> bool:
        return len({definition.property_id for definition in definitions}) == 1

    @classmethod
    def from_list_of_properties(cls, definitions: list[Property], namespace: Namespace) -> "OWLProperty":
        """Here list of properties is a list of properties with the same id, but different definitions."""

        if not cls.same_property_id(definitions):
            raise _exceptions.Error30()

        prop_dict = {
            "id_": namespace[definitions[0].property_id],
            "type_": set(),
            "label": set(),
            "comment": set(),
            "domain": set(),
            "range_": set(),
        }

        for definition in definitions:
            prop_dict["type_"].add(OWL[definition.property_type])
            prop_dict["range_"].add(
                XSD[definition.expected_value_type]
                if definition.expected_value_type in DATA_TYPE_MAPPING
                else namespace[definition.expected_value_type]
            )
            prop_dict["domain"].add(namespace[definition.class_id])

            if definition.property_name:
                prop_dict["label"].add(definition.property_name)
            if definition.description:
                prop_dict["comment"].add(definition.description)

        return cls(**prop_dict, namespace=namespace)

    # TODO: Add warnings to _exceptions.py and use them here:
    @field_validator("type_")
    def is_multi_type(cls, v, info: FieldValidationInfo):
        if len(v) > 1:
            warnings.warn(
                _exceptions.Warning30(remove_namespace(info.data["id_"]), [remove_namespace(t) for t in v]).message,
                category=_exceptions.Warning30,
                stacklevel=2,
            )
        return v

    @field_validator("range_")
    def is_multi_range(cls, v, info: FieldValidationInfo):
        if len(v) > 1:
            warnings.warn(
                _exceptions.Warning31(remove_namespace(info.data["id_"]), [remove_namespace(t) for t in v]).message,
                category=_exceptions.Warning31,
                stacklevel=2,
            )
        return v

    @field_validator("domain")
    def is_multi_domain(cls, v, info: FieldValidationInfo):
        if len(v) > 1:
            warnings.warn(
                _exceptions.Warning32(remove_namespace(info.data["id_"]), [remove_namespace(t) for t in v]).message,
                category=_exceptions.Warning32,
                stacklevel=2,
            )
        return v

    @field_validator("label")
    def has_multi_name(cls, v, info: FieldValidationInfo):
        if len(v) > 1:
            warnings.warn(
                _exceptions.Warning33(remove_namespace(info.data["id_"]), v).message,
                category=_exceptions.Warning33,
                stacklevel=2,
            )
        return v

    @field_validator("comment")
    def has_multi_comment(cls, v, info: FieldValidationInfo):
        if len(v) > 1:
            warnings.warn(
                _exceptions.Warning34(remove_namespace(info.data["id_"])).message,
                category=_exceptions.Warning34,
                stacklevel=2,
            )
        return v

    @property
    def domain_triples(self) -> list[tuple]:
        triples = []
        if len(self.domain) == 1:
            triples.append((self.id_, RDFS.domain, next(iter(self.domain))))
        else:
            _graph = Graph()
            b_union = BNode()
            b_domain = BNode()
            _graph.add((self.id_, RDFS.domain, b_domain))
            _graph.add((b_domain, OWL.unionOf, b_union))
            _graph.add((b_domain, RDF.type, OWL.Class))
            _ = GraphCollection(_graph, b_union, list(self.domain))
            triples.extend(list(_graph))
        return triples

    @property
    def range_triples(self) -> list[tuple]:
        triples = []
        if len(self.range_) == 1:
            triples.append((self.id_, RDFS.range, next(iter(self.range_))))
        else:
            _graph = Graph()
            b_union = BNode()
            b_range = BNode()
            _graph.add((self.id_, RDFS.range, b_range))
            _graph.add((b_range, OWL.unionOf, b_union))
            _graph.add((b_range, RDF.type, OWL.Class))
            _graph.add((b_range, OWL.unionOf, b_union))
            _graph.add((b_range, RDF.type, OWL.Class))
            _ = GraphCollection(_graph, b_union, list(self.range_))
            triples.extend(list(_graph))
        return triples

    @property
    def type_triples(self) -> list[tuple]:
        return [(self.id_, RDF.type, type_) for type_ in self.type_]

    @property
    def label_triples(self) -> list[tuple]:
        label = list(filter(None, self.label))
        return [(self.id_, RDFS.label, Literal(label[0] if label else self.id_))]

    @property
    def comment_triples(self) -> list[tuple]:
        return [(self.id_, RDFS.comment, Literal("\n".join(filter(None, self.comment))))]

    @property
    def triples(self) -> list[tuple]:
        return self.type_triples + self.label_triples + self.comment_triples + self.domain_triples + self.range_triples


SHACL = Namespace("http://www.w3.org/ns/shacl#")


class SHACLNodeShape(OntologyModel):
    id_: URIRef
    type_: URIRef = SHACL.NodeShape
    target_class: URIRef
    property_shapes: list["SHACLPropertyShape"]
    namespace: Namespace

    @property
    def type_triples(self) -> list[tuple]:
        return [(self.id_, RDF.type, self.type_)]

    @property
    def target_class_triples(self) -> list[tuple]:
        return [(self.id_, SHACL.targetClass, self.target_class)]

    @property
    def property_shapes_triples(self) -> list[tuple]:
        triples = []
        for property_shape in self.property_shapes:
            triples.append((self.id_, SHACL.property, property_shape.id_))
            triples.extend(property_shape.triples)
        return triples

    @property
    def triples(self) -> list[tuple]:
        return self.type_triples + self.target_class_triples + self.property_shapes_triples

    @classmethod
    def from_rules(
        cls, class_definition: Class, property_definitions: list[Property], namespace: Namespace
    ) -> "SHACLNodeShape":
        return cls(
            id_=namespace[f"{class_definition.class_id}Shape"],
            target_class=namespace[class_definition.class_id],
            property_shapes=[SHACLPropertyShape.from_property(prop, namespace) for prop in property_definitions],
            namespace=namespace,
        )


class SHACLPropertyShape(OntologyModel):
    id_: BNode
    type_: URIRef = SHACL.property
    path: URIRef  # URIRef to property in OWL
    node_kind: URIRef  # SHACL.IRI or SHACL.Literal
    expected_value_type: URIRef
    min_count: Optional[int]
    max_count: Optional[int]
    namespace: Namespace

    @property
    def path_triples(self) -> list[tuple]:
        return [(self.id_, SHACL.path, self.path)]

    @property
    def node_kind_triples(self) -> list[tuple]:
        triples = [(self.id_, SHACL.nodeKind, self.node_kind)]

        if self.node_kind == SHACL.Literal:
            triples.append((self.id_, SHACL.datatype, self.expected_value_type))
        else:
            triples.append((self.id_, SHACL.node, self.expected_value_type))

        return triples

    @property
    def cardinality_triples(self) -> list[tuple]:
        triples = []
        if self.min_count:
            triples.append((self.id_, SHACL.minCount, Literal(self.min_count)))
        if self.max_count:
            triples.append((self.id_, SHACL.maxCount, Literal(self.max_count)))

        return triples

    @property
    def triples(self) -> list[tuple]:
        return self.path_triples + self.node_kind_triples + self.cardinality_triples

    @classmethod
    def from_property(cls, definition: Property, namespace: Namespace) -> "SHACLPropertyShape":
        return cls(
            id_=BNode(),
            path=namespace[definition.property_id],
            node_kind=SHACL.IRI if definition.property_type == "ObjectProperty" else SHACL.Literal,
            expected_value_type=(
                namespace[f"{definition.expected_value_type}Shape"]
                if definition.property_type == "ObjectProperty"
                else XSD[definition.expected_value_type]
            ),
            min_count=definition.min_count,
            max_count=definition.max_count,
            namespace=namespace,
        )



"""Methods to transform Domain Knowledge Graph to App Knowledge Graph
"""

import logging
import time
import traceback
from enum import StrEnum
from typing import List, Tuple

import pandas as pd
from cognite.client import CogniteClient
from prometheus_client import Gauge
from pydantic import BaseModel
from rdflib import RDF, Graph
from rdflib.term import Literal

from cognite.neat.graph.transformations.query_generator.sparql import build_sparql_query
from cognite.neat.rules.models import TransformationRules
from cognite.neat.rules.to_rdf_path import AllProperties, AllReferences, RawLookup, parse_rule
from cognite.neat.utils.utils import remove_namespace

prom_total_proc_rules_g = Gauge("neat_total_processed_rules", "Number of processed rules", ["state"])
rules_processing_timing_metric = Gauge(
    "neat_rules_processing_timing", "Transformation rules processing timing metrics", ["aggregation"]
)


class Triple(StrEnum):
    subject = "subject"
    predicate = "predicate"
    object = "object"


COMMIT_BATCH_SIZE = 10000


class RuleProcessingReportRec(BaseModel):
    """Report record for rule processing"""

    row_id: str = None
    rule_name: str = None
    rule_type: str = None
    rule_expression: str = None
    status: str = None
    error_message: str = None
    elapsed_time: float = 0
    rows_in_response: int = 0


class RuleProcessingReport(BaseModel):
    """Report for rule processing"""

    total_rules: int = 0
    total_success: int = 0
    total_success_no_results: int = 0
    total_failed: int = 0
    records: List[RuleProcessingReportRec] = []
    elapsed_time: float = 0


def source2solution_graph(
    source_knowledge_graph: Graph,
    transformation_rules: TransformationRules,
    solution_knowledge_graph: Graph = None,
    client: CogniteClient = None,
    cdf_lookup_database: str = None,
    extra_triples: List[Tuple] = None,
    stop_on_exception: bool = False,
    missing_raw_lookup_value: str = "NaN",
    processing_report: RuleProcessingReport = None,
):
    """Create solution graph from source graph and transformation rules"""

    # TODO: This is to be improved and slowly sunset domain2app_knowledge_graph

    return domain2app_knowledge_graph(
        domain_knowledge_graph=source_knowledge_graph,
        transformation_rules=transformation_rules,
        app_instance_graph=solution_knowledge_graph,
        client=client,
        cdf_lookup_database=cdf_lookup_database,
        extra_triples=extra_triples,
        stop_on_exception=stop_on_exception,
        missing_raw_lookup_value=missing_raw_lookup_value,
        processing_report=processing_report,
    )


def domain2app_knowledge_graph(
    domain_knowledge_graph: Graph,
    transformation_rules: TransformationRules,
    app_instance_graph: Graph = None,
    client: CogniteClient = None,
    cdf_lookup_database: str = None,
    extra_triples: List[Tuple] = None,
    stop_on_exception: bool = False,
    missing_raw_lookup_value: str = "NaN",
    processing_report: RuleProcessingReport = None,
) -> Graph:
    """Generates App specific knowledge graph based on Domain Knowledge Graph

    Parameters
    ----------
    domain_knowledge_graph : Graph
        _description_


    Returns
    -------
    Graph
        _description_
    """
    if app_instance_graph is None:
        app_instance_graph = Graph()
        # Bind App namespace and prefix
        app_instance_graph.bind(transformation_rules.metadata.prefix, transformation_rules.metadata.namespace)
        # Bind other prefixes and namespaces
        for prefix, namespace in transformation_rules.prefixes.items():
            app_instance_graph.bind(prefix, namespace)

    tables_by_name = {}
    for table_name in transformation_rules.raw_tables:
        logging.debug(f"Loading {table_name} table from database {cdf_lookup_database}")
        table = client.raw.rows.retrieve_dataframe(cdf_lookup_database, table_name, limit=-1)
        tables_by_name[table_name] = table

    # Add references with their type first
    types = []
    success = 0
    success_no_results = 0
    failed = 0
    commit_counter = 0
    timing_traces = []
    prom_total_proc_rules_g.labels(state="success_no_results").set(0)
    prom_total_proc_rules_g.labels(state="success").set(0)
    prom_total_proc_rules_g.labels(state="failed").set(0)

    def check_commit(force_commit: bool = False):
        """'Commit nodes to the graph if batch counter is reached or if force_commit is True"""

        if force_commit:
            logging.debug("Committing nodes")
            app_instance_graph.commit()
            logging.debug("Nodes committed")
            return
        nonlocal commit_counter
        commit_counter += 1
        if commit_counter >= COMMIT_BATCH_SIZE:
            logging.info(f"Committing {COMMIT_BATCH_SIZE} nodes")
            app_instance_graph.commit()
            logging.info(f" {COMMIT_BATCH_SIZE} nodes committed")
            commit_counter = 0

    proc_start_time = time.perf_counter()
    for sheet_row, rule_definition in transformation_rules.properties.items():
        if not rule_definition.rule or rule_definition.skip_rule:
            continue
        msg = f"Processing {sheet_row}: class <{rule_definition.class_id}> "
        msg += f"property <{rule_definition.property_name}> rule <{rule_definition.rule}>"

        processing_report_rec = RuleProcessingReportRec(
            row_id=sheet_row,
            rule_name=f"{rule_definition.class_id}_{rule_definition.property_name}",
            rule_type=rule_definition.rule_type,
            rule_expression=rule_definition.rule,
        )
        logging.info(msg)
        try:
            start_time = time.perf_counter()
            # Parse rule:
            rule = parse_rule(rule_definition.rule, rule_definition.rule_type)

            # Build SPARQL if needed:
            if rule_definition.rule_type == "sparql":
                query = rule.traversal.query
            else:
                query = build_sparql_query(domain_knowledge_graph, rule.traversal, transformation_rules.prefixes)

            logging.info(f"Query: {query}")

            if query_results := list(domain_knowledge_graph.query(query)):
                # Generate URI for class and property in target namespace
                class_URI = transformation_rules.metadata.namespace[rule_definition.class_id]
                property_URI = transformation_rules.metadata.namespace[rule_definition.property_name]

                # Turn query results into dataframe
                instance_df = pd.DataFrame(query_results, columns=list(Triple))

                # If we are not grabbing all properties for class instances
                # then we are able to replace source property URI with target property URI
                # otherwise we should keep source property URI
                if not isinstance(rule.traversal, AllProperties):
                    instance_df[Triple.predicate] = property_URI

                # If we are storing object from the source graph as literal value(propety type being Datatype Property)
                # in the target graph then we should remove namespace from the object URI and store it as literal
                if isinstance(rule.traversal, AllReferences) and rule_definition.property_type == "DatatypeProperty":
                    instance_df[Triple.object] = instance_df[Triple.object].apply(
                        lambda x: Literal(remove_namespace(x))
                    )

                if isinstance(rule, RawLookup):
                    lookup_map = tables_by_name[rule.table.name].set_index(rule.table.key)[rule.table.value].to_dict()

                    def lookup(
                        literal: Literal,
                        lookup_table=lookup_map,
                        missing_raw_lookup_value=missing_raw_lookup_value,
                    ):
                        if new_value := lookup_table.get(literal.value):
                            return Literal(new_value, literal.language, literal.datatype, literal.normalize)
                        elif missing_raw_lookup_value:
                            return Literal(
                                missing_raw_lookup_value, literal.language, literal.datatype, literal.normalize
                            )
                        else:
                            return literal

                    instance_df[Triple.object] = instance_df[Triple.object].apply(lookup)

                # Add instances
                for _, triple in instance_df.iterrows():
                    app_instance_graph.add(triple.values)
                    check_commit()
                # Setting instances type and merging them with df containing instance - type relations
                instance_df[Triple.predicate] = RDF.type
                instance_df[Triple.object] = class_URI
                types.append(instance_df)
                success += 1
                prom_total_proc_rules_g.labels(state="success").inc()
                elapsed_time = time.perf_counter() - start_time
                timing_traces.append(elapsed_time)
                processing_report_rec.elapsed_time = elapsed_time
                processing_report_rec.status = "success"
                processing_report_rec.rows_in_response = len(instance_df)
            else:
                success_no_results += 1
                prom_total_proc_rules_g.labels(state="success_no_results").inc()
                elapsed_time = time.perf_counter() - start_time
                timing_traces.append(elapsed_time)
                processing_report_rec.elapsed_time = elapsed_time
                processing_report_rec.status = "success_no_results"

        except Exception as e:
            failed += 1
            elapsed_time = time.perf_counter() - start_time
            processing_report_rec.elapsed_time = elapsed_time
            processing_report_rec.status = "failed"
            processing_report_rec.error_message = str(e)
            prom_total_proc_rules_g.labels(state="failed").inc()
            logging.error(
                f" Error while processing rule {rule_definition.rule} for class {rule_definition.class_id} \
                and property {rule_definition.property_name}"
            )
            logging.error(traceback.format_exc())
            if stop_on_exception:
                raise e

        if processing_report:
            processing_report.records.append(processing_report_rec)

    if processing_report:
        processing_report.total_rules = len(transformation_rules.properties)
        processing_report.total_success = success
        processing_report.total_success_no_results = success_no_results
        processing_report.total_failed = failed
        processing_report.elapsed_time = time.perf_counter() - proc_start_time

    if timing_traces:
        df = pd.Series(timing_traces)
        rules_processing_timing_metric.labels(aggregation="sum").set(df.sum())
        rules_processing_timing_metric.labels(aggregation="std_div").set(df.std())
        rules_processing_timing_metric.labels(aggregation="min").set(df.min())
        rules_processing_timing_metric.labels(aggregation="max").set(df.max())
        rules_processing_timing_metric.labels(aggregation="mean").set(df.mean())

    type_df = pd.concat(types).drop_duplicates(Triple.subject).reset_index(drop=True)

    # Add instance - RDF Type relations
    for _, triple in type_df.iterrows():
        app_instance_graph.add(triple.values)
        check_commit()

    if extra_triples:
        for i, triple in enumerate(extra_triples):
            try:
                app_instance_graph.add(triple)
                check_commit()
            except ValueError:
                raise ValueError(f"Triple {i} in extra_triples is not correct and cannot be added!")

    check_commit(force_commit=True)
    return app_instance_graph



from pathlib import Path

from rdflib import Graph, Namespace

from cognite.neat.constants import DEFAULT_NAMESPACE, DEFAULT_URI, PREFIXES

__all__ = ["rdf_file_to_graph"]


def rdf_file_to_graph(
    filepath: Path,
    base_prefix: str = DEFAULT_URI,
    base_namespace: Namespace = DEFAULT_NAMESPACE,
    prefixes: dict[str, Namespace] = PREFIXES,
) -> Graph:
    """Loads RDF from file

    Parameters
    ----------
    filepath : Path
        File path to RDF
    namespace : Namespace, optional
        The logical URI to use as the graph namespace base, by default "http://purl.org/cognite/app-dm"
    prefixes : dict, optional
        Dictionary containing prefix-namespace pairs, by default PREFIXES

    Returns
    -------
    Graph
        An RDF Graph
    """
    graph = Graph()
    if filepath.is_file():
        graph.parse(filepath, publicID=base_namespace)
    else:
        for filename in filepath.iterdir():
            if filename.is_file():
                graph.parse(filename, publicID=base_namespace)

    graph.bind(base_prefix, base_namespace)
    if prefixes:
        for prefix, namespace in prefixes.items():
            graph.bind(prefix, namespace)

    return graph



# IO, CDF, Workflows


import logging
import os
import shutil
from fastapi import APIRouter, UploadFile
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigItem

from cognite.neat.workflows.utils import get_file_hash
from cognite.neat.app.api.configuration import neat_app

router = APIRouter()


@router.get("/api/cdf/neat-resources")
def get_neat_resources(resource_type: str = None):
    result = neat_app.cdf_store.get_list_of_resources_from_cdf(resource_type=resource_type)
    logging.debug(f"Got {len(result)} resources")
    return {"result": result}


@router.post("/api/cdf/init-neat-resources")
def init_neat_cdf_resources(resource_type: str = None):
    neat_app.cdf_store.init_cdf_resources(resource_type=resource_type)
    return {"result": "ok"}


@router.post("/api/file/upload/{workflow_name}/{file_type}/{step_id}/{action}")
async def file_upload_handler(files: list[UploadFile], workflow_name: str, file_type: str, step_id: str, action: str):
    # get directory path
    upload_dir = neat_app.config.rules_store_path
    file_name = ""
    file_version = ""
    for file in files:
        logging.info(
            f"Uploading file : {file.filename} , workflow : {workflow_name} , step_id {step_id} , action : {action}"
        )
        # save file to disk
        full_path = os.path.join(upload_dir, file.filename)
        with open(full_path, "wb") as buffer:
            shutil.copyfileobj(file.file, buffer)
        file_name = file.filename
        file_version = get_file_hash(full_path)
        break  # only one file is supported for now

    if "update_config" in action and file_type == "rules":
        logging.info("Automatically updating workflow config")
        workflow = neat_app.workflow_manager.get_workflow(workflow_name)
        workflow_definition = workflow.get_workflow_definition()

        # update config item rules.file with the new file name
        config_item = workflow_definition.get_config_item("rules.file")
        if config_item is None:
            config_item = WorkflowConfigItem(name="rules.file", value=file_name, label="Rules file name", group="rules")
        config_item.value = file_name
        workflow_definition.upsert_config_item(config_item)
        # update config item rules.file with the new file name
        config_item = workflow_definition.get_config_item("rules.version")
        if config_item is None:
            config_item = WorkflowConfigItem(name="rules.version", value="", label="Rules file version", group="rules")
            workflow_definition.upsert_config_item(config_item)
        neat_app.workflow_manager.save_workflow_to_storage(workflow_name)

    if "start_workflow" in action:
        logging.info("Starting workflow after file upload")
        workflow = neat_app.workflow_manager.get_workflow(workflow_name)
        flow_msg = FlowMessage(
            payload={"file_name": file_name, "hash": file_version, "full_path": full_path, "file_type": file_type}
        )
        start_step_id = None if step_id == "none" else step_id

        workflow.start(sync=False, flow_message=flow_msg, start_step_id=start_step_id)

    return {"file_name": file_name, "hash": file_version}



"""Should contain methods to validate Graph Transformation Rules sheet,
as well App Data Model (RDF)
"""

import logging
from typing import List, Tuple

from cognite.client.data_classes import Asset


def _find_circular_reference_path(asset: Asset, assets: dict[str, Asset], max_hierarchy_depth: int = 10000) -> List:
    original_external_id = asset.get("external_id")
    circle = [original_external_id]
    ref = assets.get(asset.get("parent_external_id"))

    hop = 0
    while ref is not None and hop < max_hierarchy_depth:
        hop += 1
        circle.append(ref.get("external_id"))
        if len(circle) != len(set(circle)):
            msg = (
                f"Found circular reference in asset hierarchy which starts with "
                f"{original_external_id} and enters loop at {circle[-1]}. "
            )
            logging.error(msg)
            return circle

        ref = assets.get(ref.get("parent_external_id"))

    if hop >= max_hierarchy_depth:
        msg = (
            f"Your asset hierarchy is too deep. Max depth is {max_hierarchy_depth}. "
            "You probably have a circular reference."
        )
        logging.error(msg)
        return circle
    else:
        return []


def validate_asset_hierarchy(assets: dict[str, Asset]) -> Tuple[List[str], List[List[str]]]:
    """Validates asset hierarchy and reports on orphan assets and circular dependency

    Parameters
    ----------
    assets : dict[str, Asset]
        A dictionary of assets with external_id as key

    Returns
    -------
    Tuple[List[Asset], List[List[str]]
        List of orphan assets external ids and list of circular path of external ids.
        If both lists are empty, the hierarchy is healthy.
    """
    orphan_assets = []
    circular_reference_paths = []
    for asset in assets.values():
        parent_external_id = asset.get("parent_external_id")
        if parent_external_id is not None and parent_external_id not in assets:
            msg = (
                f"Found orphan asset {asset.get('external_id')} with parent {parent_external_id} which does not exist."
            )
            logging.error(msg)
            orphan_assets.append(asset.get("external_id"))
        circular_reference_path = _find_circular_reference_path(asset, assets)
        if not len(circular_reference_path):
            continue

        # Save the circle only once, not once for every asset
        if set(circular_reference_path) in [set(path) for path in circular_reference_paths]:
            continue
        circular_reference_paths.append(circular_reference_path)
    return orphan_assets, circular_reference_paths



from datetime import datetime, timezone
import re
from typing import Any
from typing_extensions import TypeAliasType
import warnings
from pydantic import BaseModel, ConfigDict, Field, create_model
from pydantic._internal._model_construction import ModelMetaclass
from rdflib import Graph, URIRef

from cognite.client.data_classes import Asset, Relationship
from cognite.neat.graph.loaders.core.rdf_to_assets import NeatMetadataKeys
from cognite.neat.rules.analysis import (
    to_class_property_pairs,
    define_class_asset_mapping,
    define_class_relationship_mapping,
)
from cognite.neat.graph.transformations.query_generator.sparql import build_construct_query, triples2dictionary
from cognite.neat.rules.models import Property, TransformationRules, type_to_target_convention
from cognite.neat.rules import _exceptions

EdgeOneToOne = TypeAliasType("EdgeOneToOne", str)
EdgeOneToMany = TypeAliasType("EdgeOneToMany", list[str])


def default_model_configuration():
    return ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )


def default_model_methods():
    return [from_graph, to_asset, to_relationship, to_dms, to_graph]


def rules_to_pydantic_models(
    transformation_rules: TransformationRules, methods: list = None
) -> dict[str, ModelMetaclass]:
    """Generate pydantic models from transformation rules.

    Parameters
    ----------
    transformation_rules : TransformationRules
        Transformation rules
    methods : list, optional
        List of methods to register for pydantic models , by default None

    Returns
    -------
    dict[str, ModelMetaclass]
        Dictionary containing pydantic models

    Notes
    -----
    Currently this will take only unique properties and those which column rule_type
    is set to rdfpath, hence only_rdfpath = True. This means that at the moment
    we do not support UNION, i.e. ability to handle multiple rdfpaths for the same
    property. This is needed option and should be added in the second version of the exporter.
    """
    if methods is None:
        methods = default_model_methods()

    class_property_pairs = to_class_property_pairs(transformation_rules, only_rdfpath=True)

    models: dict[str, ModelMetaclass] = {}
    for class_, properties in class_property_pairs.items():
        # generate fields from define properties
        fields = _properties_to_pydantic_fields(properties)

        # store default class to relationship mapping field
        # which is used by the `to_relationship` method
        fields["class_to_asset_mapping"] = (
            dict[str, list[str]],
            Field(define_class_asset_mapping(transformation_rules, class_)),
        )

        # store default class to relationship mapping field
        # which is used by the `to_relationship` method
        fields["class_to_relationship_mapping"] = (
            dict[str, list[str]],
            Field(define_class_relationship_mapping(transformation_rules, class_)),
        )

        fields["data_set_id"] = (
            int,
            Field(
                transformation_rules.metadata.data_set_id or None,
            ),
        )

        model = _dictionary_to_pydantic_model(
            class_, fields, methods=[from_graph, to_asset, to_relationship, to_dms, to_graph]
        )

        models[class_] = model

    return models


def _properties_to_pydantic_fields(
    properties: dict[str, Property],
) -> dict[str, tuple[EdgeOneToMany | EdgeOneToOne | type | list[type], Any]]:
    """Turns definition of properties into pydantic fields.

    Parameters
    ----------
    properties : dict[str, Property]
        Dictionary of properties

    Returns
    -------
    dict[str, tuple[EdgeOneToMany | EdgeOneToOne | type | list[type], Any]]
        Dictionary of pydantic fields
    """

    fields: dict[str, tuple[EdgeOneToMany | EdgeOneToOne | type | list[type], Any]] = {}

    fields = {"external_id": (str, Field(..., alias="external_id"))}

    for name, property_ in properties.items():
        field_type = _define_field_type(property_)

        field_definition: dict = {"alias": name}

        if field_type.__name__ in [EdgeOneToMany.__name__, list.__name__]:
            field_definition["min_length"] = property_.min_count
            field_definition["max_length"] = property_.max_count

        if not property_.is_mandatory and not property_.default:
            field_definition["default"] = None
        elif property_.default:
            field_definition["default"] = property_.default

        # making sure that field names are python compliant
        # their original names are stored as aliases
        fields[re.sub(r"[^_a-zA-Z0-9/_]", "_", name)] = (field_type, Field(**field_definition))

    return fields


def _define_field_type(property_: Property):
    if property_.property_type == "ObjectProperty" and property_.max_count == 1:
        return EdgeOneToOne
    elif property_.property_type == "ObjectProperty":
        return EdgeOneToMany
    elif property_.property_type == "DatatypeProperty" and property_.max_count == 1:
        return type_to_target_convention(property_.expected_value_type, "python")
    else:
        return list[type_to_target_convention(property_.expected_value_type, "python")]


def _dictionary_to_pydantic_model(
    name: str,
    model_definition: dict,
    model_configuration: ConfigDict = None,
    methods: list = None,
    validators: list = None,
) -> type[BaseModel]:
    """Generates pydantic model from dictionary containing definition of fields.
    Additionally it adds methods to the model and validators.

    Parameters
    ----------
    name : str
        Name of the model
    model_definition : dict
        Dictionary containing definition of fields
    model_configuration : ConfigDict, optional
        Configuration of pydantic model, by default None
    methods : list, optional
        Methods that work on fields once model is instantiated, by default None
    validators : list, optional
        Any custom validators to be added in addition to base pydantic ones, by default None

    Returns
    -------
    type[BaseModel]
        Pydantic model
    """

    if model_configuration:
        model_configuration = default_model_configuration()

    fields = {}

    for field_name, value in model_definition.items():
        if isinstance(value, tuple):
            fields[field_name] = value
        elif isinstance(value, dict):
            fields[field_name] = (_dictionary_to_pydantic_model(f"{name}_{field_name}", value), ...)
        else:
            raise _exceptions.Error40(field_name, value)

    model = create_model(name, __config__=model_configuration, **fields)

    if methods:
        for method in methods:
            setattr(model, method.__name__, method)

    # any additional validators to be added
    if validators:
        ...

    return model


# Define methods that work on model instance
@classmethod
def from_graph(cls, graph: Graph, transformation_rules: TransformationRules, external_id: URIRef):
    """Method that creates model instance from class instance stored in graph.

    Parameters
    ----------
    graph : Graph
        Graph containing triples of class instance
    transformation_rules : TransformationRules
        Transformation rules
    external_id : URIRef
        External id of class instance to be used to instantiate associated pydantic model

    Returns
    -------
    cls
        Pydantic model instance
    """
    # build sparql query for given object
    class_ = cls.__name__
    sparql_query = build_construct_query(
        graph, class_, transformation_rules, class_instances=[external_id], properties_optional=False
    )

    result = triples2dictionary(list(graph.query(sparql_query)))

    # wrangle results to dict
    for field in cls.model_fields.values():
        # if field is not required and not in result, skip
        if not field.is_required() and field.alias not in result:
            continue

        # if field is required and not in result, raise error
        if field.is_required() and field.alias not in result:
            raise _exceptions.Error41(field.alias, external_id)

        # flatten result if field is not edge or list of values
        if field.annotation.__name__ not in [EdgeOneToMany.__name__, list.__name__]:
            if isinstance(result[field.alias], list) and len(result[field.alias]) > 1:
                warnings.warn(
                    _exceptions.Warning41(
                        field.alias,
                        len(result[field.alias]),
                    ).message,
                    category=_exceptions.Warning41,
                    stacklevel=2,
                )

            result[field.alias] = result[field.alias][0]

    return cls(**result)


# define methods that creates asset out of model id (default)
def to_asset(
    self,
    add_system_metadata: bool = True,
    metadata_keys: NeatMetadataKeys | None = None,
    add_labels: bool = True,
    data_set_id: int = None,
) -> Asset:
    """Convert model instance to asset.

    Parameters
    ----------
    add_system_metadata : bool, optional
        Flag indicating to add or not system/neat metadata, by default True
    metadata_keys : NeatMetadataKeys | None, optional
        Definition of system/neat metadata, by default None
    add_labels : bool, optional
        To add or not labels to asset, by default True
    data_set_id : int, optional
        Data set id to which asset belongs to, by default None

    Returns
    -------
    Asset
        Asset instance
    """
    # Needs copy otherwise modifications impact all instances
    class_instance_dictionary = self.model_dump(by_alias=True)
    adapted_mapping_config = _adapt_mapping_config_by_instance(
        self.external_id, class_instance_dictionary, self.class_to_asset_mapping
    )
    asset = _class_to_asset_instance_dictionary(class_instance_dictionary, adapted_mapping_config)

    # set default metadata keys if not provided
    metadata_keys = NeatMetadataKeys() if metadata_keys is None else metadata_keys

    # add system metadata
    if add_system_metadata:
        _add_system_metadata(self, metadata_keys, asset)

    if add_labels:
        asset["labels"] = [asset["metadata"][metadata_keys.type], "non-historic"]

    if data_set_id:
        return Asset(**asset, data_set_id=data_set_id)
    else:
        return Asset(**asset, data_set_id=self.data_set_id)


def _add_system_metadata(self, metadata_keys: NeatMetadataKeys, asset: dict):
    asset["metadata"][metadata_keys.type] = self.__class__.__name__
    asset["metadata"][metadata_keys.identifier] = self.external_id
    now = str(datetime.now(timezone.utc))
    asset["metadata"][metadata_keys.start_time] = now
    asset["metadata"][metadata_keys.update_time] = now
    asset["metadata"][metadata_keys.active] = "true"


def _adapt_mapping_config_by_instance(external_id, class_instance_dictionary, mapping_config):
    adapted_mapping_config = {}
    for asset_field, class_properties in mapping_config.items():
        if asset_field != "metadata":
            # We are selecting first property that is available in the graph
            # and add it to corresponding asset field and exit loop
            for property_ in class_properties:
                if class_instance_dictionary.get(property_, None):
                    adapted_mapping_config[asset_field] = property_
                    break

        elif metadata_keys := [
            property_ for property_ in class_properties if class_instance_dictionary.get(property_, None)
        ]:
            adapted_mapping_config["metadata"] = metadata_keys

    # Raise warnings for fields that will miss in asset
    for asset_field in mapping_config:
        if asset_field not in adapted_mapping_config:
            warnings.warn(
                _exceptions.Warning40(external_id, asset_field).message,
                category=_exceptions.Warning40,
                stacklevel=2,
            )

    return adapted_mapping_config


def _class_to_asset_instance_dictionary(class_instance_dictionary, mapping_config):
    for key, values in mapping_config.items():
        if key != "metadata":
            mapping_config[key] = class_instance_dictionary.get(values, None)
        else:
            mapping_config[key] = {value: class_instance_dictionary.get(value, None) for value in values}

    return mapping_config


def to_relationship(self, transformation_rules: TransformationRules) -> Relationship:
    """Creates relationship instance from model instance."""
    ...


def to_dms(self, transformation_rules: TransformationRules):
    """Creates instance of dm in CDF."""
    ...


def to_graph(self, transformation_rules: TransformationRules, graph: Graph):
    """Writes instance as set of triples to triple store (Graphs)."""
    ...



from abc import ABC, abstractmethod
from typing import ClassVar, TypeVar
from pydantic import BaseModel, ConfigDict
from cognite.neat.app.monitoring.metrics import NeatMetricsCollector

from cognite.neat.workflows.model import WorkflowConfigItem, WorkflowConfigs


class Config(BaseModel):
    ...


class DataContract(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)


T_Input = TypeVar("T_Input", bound=DataContract)

T_Output = TypeVar("T_Output", bound=DataContract)


class Step(ABC):
    description: str = ""
    category: str = "default"
    configuration_templates: list[WorkflowConfigItem] = []
    scope: str = "global"
    metrics: NeatMetricsCollector | None = None
    configs: WorkflowConfigs | None = None

    def __init__(self, data_store_path: str | None = None):
        self.log: bool = False
        self.data_store_path: str = data_store_path

    def set_metrics(self, metrics: NeatMetricsCollector):
        self.metrics = metrics

    def set_workflow_configs(self, configs: WorkflowConfigs):
        self.configs = configs

    def set_flow_context(self, context: dict[str, DataContract]):
        self.flow_context = context

    @abstractmethod
    def run(self, *input_data: T_Input) -> T_Output:
        ...



import logging
from pathlib import Path
from cognite.neat.constants import PREFIXES

from cognite.neat.graph.stores import RdfStoreType
from cognite.neat.graph.stores import NeatGraphStore, drop_graph_store
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigItem
from cognite.neat.workflows.steps.step_model import Step

from ..data_contracts import RulesData, SolutionGraph, SourceGraph

__all__ = ["ConfigureDefaultGraphStores", "LoadInstancesFromRdfFileToSourceGraph", "ResetGraphStores"]


class ConfigureDefaultGraphStores(Step):
    description = "The step initializes the source and solution graph stores."
    category = "graph_store"
    configuration_templates = [
        WorkflowConfigItem(
            name="source_rdf_store.type",
            value=RdfStoreType.OXIGRAPH,
            label="Data store type for source graph. Supported: oxigraph, memory,file, graphdb, sparql. ",
        ),
        WorkflowConfigItem(
            name="solution_rdf_store.type",
            value=RdfStoreType.OXIGRAPH,
            label="Data store type for solutioin graph. Supported: oxigraph, memory,file, graphdb, sparql",
        ),
        WorkflowConfigItem(
            name="source_rdf_store.disk_store_dir",
            value="source-graph-store",
            label="Local directory for source graph store",
        ),
        WorkflowConfigItem(
            name="solution_rdf_store.disk_store_dir",
            value="solution-graph-store",
            label="Local directory for solution graph store",
        ),
        WorkflowConfigItem(
            name="stores_to_configure",
            value="all",
            label="Defines which stores to configure. Possible values: all, source, solution",
        ),
        WorkflowConfigItem(
            name="solution_rdf_store.api_root_url",
            value="",
            label="Root url for graphdb or sparql endpoint",
        ),
    ]

    def run(self, rules_data: RulesData) -> FlowMessage | SourceGraph | SolutionGraph:
        logging.info("Initializing source graph")
        stores_to_configure = self.configs.get_config_item_value("stores_to_configure", "all")
        source_store_dir = self.configs.get_config_item_value("source_rdf_store.disk_store_dir", "source_graph")
        source_store_dir = Path(self.data_store_path) / Path(source_store_dir) if source_store_dir else None
        source_store_type = self.configs.get_config_item_value("source_rdf_store.type", RdfStoreType.MEMORY)
        if stores_to_configure in ["all", "source"]:
            if source_store_type == RdfStoreType.OXIGRAPH and "SourceGraph" in self.flow_context:
                return FlowMessage(output_text="Stores already configured")

            source_graph = NeatGraphStore(
                prefixes=rules_data.rules.prefixes, base_prefix="neat", namespace=PREFIXES["neat"]
            )
            source_graph.init_graph(
                source_store_type,
                self.configs.get_config_item_value("source_rdf_store.query_url", ""),
                self.configs.get_config_item_value("source_rdf_store.update_url", ""),
                "neat-tnt",
                internal_storage_dir=source_store_dir,
            )
            if stores_to_configure == "source":
                return FlowMessage(output_text="Source graph store configured successfully"), SourceGraph(
                    graph=source_graph
                )

        if stores_to_configure in ["all", "solution"]:
            solution_store_dir = self.configs.get_config_item_value(
                "solution_rdf_store.disk_store_dir", "solution_graph"
            )
            solution_store_dir = Path(self.data_store_path) / Path(solution_store_dir) if solution_store_dir else None
            solution_store_type = self.configs.get_config_item_value("solution_rdf_store.type", RdfStoreType.MEMORY)

            if solution_store_type == RdfStoreType.OXIGRAPH and "SolutionGraph" in self.flow_context:
                return FlowMessage(output_text="Stores already configured")
            solution_graph = NeatGraphStore(
                prefixes=rules_data.rules.prefixes, base_prefix="neat", namespace=PREFIXES["neat"]
            )

            solution_graph.init_graph(
                solution_store_type,
                self.configs.get_config_item_value("solution_rdf_store.query_url", ""),
                self.configs.get_config_item_value("solution_rdf_store.update_url", ""),
                "tnt-solution",
                internal_storage_dir=solution_store_dir,
            )

            solution_graph.graph_db_rest_url = self.configs.get_config_item_value("solution_rdf_store.api_root_url", "")
            if stores_to_configure == "solution":
                return FlowMessage(output_text="Solution graph store configured successfully"), SolutionGraph(
                    graph=solution_graph
                )

        return (
            FlowMessage(output_text="All graph stores configured successfully"),
            SourceGraph(graph=source_graph),
            SolutionGraph(graph=solution_graph),
        )


class ResetGraphStores(Step):
    description = "The step resets graph stores to their initial state (clears all data)."
    category = "graph_store"

    def run(self) -> FlowMessage:
        source_store_type = self.configs.get_config_item_value("source_rdf_store.type", RdfStoreType.MEMORY)
        solution_store_type = self.configs.get_config_item_value("solution_rdf_store.type", RdfStoreType.MEMORY)
        if source_store_type == RdfStoreType.OXIGRAPH and solution_store_type == RdfStoreType.OXIGRAPH:
            if "SourceGraph" not in self.flow_context or "SolutionGraph" not in self.flow_context:
                source_store_dir = self.configs.get_config_item_value("source_rdf_store.disk_store_dir", "source_graph")
                solution_store_dir = self.configs.get_config_item_value(
                    "solution_rdf_store.disk_store_dir", "solution_graph"
                )
                source_store_dir = Path(self.data_store_path) / Path(source_store_dir) if source_store_dir else None
                solution_store_dir = (
                    Path(self.data_store_path) / Path(solution_store_dir) if solution_store_dir else None
                )
                if solution_store_dir:
                    drop_graph_store(None, Path(source_store_dir), force=True)
                if solution_store_dir:
                    drop_graph_store(None, Path(solution_store_dir), force=True)
            else:
                if "SourceGraph" in self.flow_context:
                    self.flow_context["SourceGraph"].graph.drop()
                if "SolutionGraph" in self.flow_context:
                    self.flow_context["SolutionGraph"].graph.drop()
        return FlowMessage(output_text="Stores Reset")


class LoadInstancesFromRdfFileToSourceGraph(Step):
    description = "The step loads instances from a file into the source graph.The file must be in RDF format."
    category = "graph_loader"
    configuration_templates = [
        WorkflowConfigItem(
            name="source_rdf_store.file",
            value="source-graphs/source-graph-dump.xml",
            label="File name of source graph data dump in RDF format",
        )
    ]

    def run(self, rules: RulesData, source_graph: SourceGraph) -> FlowMessage:
        if source_graph.graph.rdf_store_type.lower() in ("memory", "oxigraph"):
            if source_file := self.configs.get_config_item_value("source_rdf_store.file"):
                source_graph.graph.import_from_file(Path(self.data_store_path) / Path(source_file))
                logging.info(f"Loaded {source_file} into source graph.")
            else:
                raise ValueError("You need a source_rdf_store.file specified for source_rdf_store.type=memory")
        else:
            raise NotImplementedError(f"Graph type {source_graph.graph.rdf_store_type} is not supported.")

        return FlowMessage(output_text="Instances loaded to source graph")



import logging
import traceback

from cognite.client import CogniteClient
from cognite.client.data_classes import LabelDefinition

from cognite.neat.rules.exporter.core import get_labels
from cognite.neat.rules.models import TransformationRules


def upload_labels(
    client: CogniteClient,
    transformation_rules: TransformationRules,
    extra_labels: list = None,
    stop_on_exception: bool = False,
):
    """Upload labels to CDF

    Parameters
    ----------
    client : CogniteClient
        Instance of CogniteClient
    transformation_rules : TransformationRules
        Instance of TransformationRules which contains the labels to upload
    extra_labels : list[str], optional
        Any additional labels not defined in TransformationRules , by default ["historic", "non-historic"]
    stop_on_exception : bool, optional
        If this function fails to stop the process, by default False
    """
    if extra_labels is None:
        extra_labels = []

    try:
        logging.debug("Fetching existing labels from CDF")
        # This is to list all labels, not just the ones in the dataset since labels
        # must have unique external_ids, fixing issue that Daniel encountered
        if existing_labels := client.labels.list(limit=-1):
            existing_labels = set(existing_labels.to_pandas().external_id)
        else:
            existing_labels = {}
        logging.debug(f"Found {len(existing_labels)} existing labels in CDF")
    except Exception as e:
        logging.debug("Error fetching existing labels from CDF, no ")
        traceback.print_exc()
        if stop_on_exception:
            raise e
        existing_labels = {}

    non_existing_labels = set(list(get_labels(transformation_rules)) + extra_labels).difference(existing_labels)

    labels = []
    if non_existing_labels:
        logging.debug(f"Creating total of {len(existing_labels)} new labels in CDF")
        labels = [
            LabelDefinition(external_id=label, name=label, data_set_id=transformation_rules.metadata.data_set_id)
            for label in non_existing_labels
        ]
        res = client.labels.create(labels)
        logging.debug(f"Created total of {len(res)} new labels in CDF")
    else:
        logging.debug("No new labels created in CDF")
